import bluebelt.helpers.defaults as defaults

# main plot | pyplot.plot
plot = {
    'marker': None,
    'zorder': 20,
    'linestyle': '-',
    'linewidth': 1,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
    'size': defaults.medium
}



import pandas as pd
import matplotlib.dates as mdates

def axisformat(ax, series):

    if isinstance(series.index, pd.DatetimeIndex):
        rng = series.index.max() - series.index.min()

        if rng > pd.Timedelta('365 days'):
            ax.xaxis.set_major_locator(mdates.MonthLocator(interval=3))
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%m-%Y'))
            ax.set_xlabel('month')
        elif rng > pd.Timedelta('183 days'):
            ax.xaxis.set_major_locator(mdates.MonthLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%b'))
            ax.set_xlabel('month')
        elif rng > pd.Timedelta('31 days'):
            ax.xaxis.set_major_locator(mdates.WeekdayLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%V'))
            ax.set_xlabel('week')
        else:
            ax.xaxis.set_major_locator(mdates.DayLocator(interval=7))
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))
            ax.set_xlabel('date')
    else:
        ax.set_xlabel(series.name)

    return ax



import bluebelt.helpers.defaults as defaults

# main plot | pyplot.plot
scatter = {
    'linewidth': 0,
    'marker': 'o',
    'zorder': 50,
}

# histogram | pyplot.hist
histogram = {
    'density': True, 
    'alpha': 1, # alpha fraction 0-1
    'edgecolor': None, # color name, hex or None
    'facecolor': defaults.blue, # color name, hex or None
    'fill': False, #True, False
    'hatch': '//', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
    'linestyle': 'solid', # '-', '--', '-.', ':', '', (offset, on-off-seq), ...
    'linewidth': 1, # float
}

# standard deviation text | pyplot.text
text = {
    'backgroundcolor': defaults.white+(0.3,),
    'va': 'top',
    'ha': 'left',
    'fontsize': defaults.xsmall
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import pandas as pd
import numpy as np

def get_subgroups(series, subgroups=None, subgroup_size=None):
    if subgroups is not None:
        subgroup_size = subgroups.value_counts().max()
        s = pd.Series(index=subgroups, data=series.values)
        groups = [(s[group].append(pd.Series((subgroup_size - len(s[group])) * [np.nan], dtype=float), ignore_index=True)) for group in np.unique(subgroups)]
        return pd.DataFrame(groups).T
    elif subgroup_size is not None:
        series = series.append(pd.Series(((subgroup_size - series.size) % subgroup_size) * [np.NaN], dtype=float), ignore_index=True)
        return pd.DataFrame(series.values.reshape(subgroup_size, int(series.size/subgroup_size)))
    else:
        return series


import numpy as np

def mape(forecast, actuals):
    return (np.abs((actuals - forecast)/actuals).sum()) / len(forecast)

def smape(forecast, actuals):
    return (np.abs(actuals - forecast) / ((np.abs(actuals) + np.abs(forecast)) / 2) ).sum() / len(forecast)

def mda(forecast, actuals):
    return (((forecast < forecast.shift(-1)).iloc[:-1] == (actuals < actuals.shift(-1)).iloc[:-1]) * 1).sum() / (len(forecast) - 1)



import pandas as pd
import numpy as np
import scipy.stats as stats
import bluebelt.data.subsets

import bluebelt.datetime.dt

import bluebelt.statistics.std

import bluebelt.analysis.patterns
import bluebelt.statistics.hypothesis_testing

import bluebelt.core.decorators

@bluebelt.core.decorators.class_methods
@pd.api.extensions.register_dataframe_accessor('blue')
class DataFrameToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.datetime = self.datetime(self._obj)
        self.std = self.std(self._obj)
        self.patterns = self.patterns(self._obj)
        self.statistics = self.statistics(self._obj)
        self.test = self.test(self._obj)

    # first level functions

    # subsets
    def subset(self, inverse=False, **kwargs):
        return bluebelt.data.subsets.subset(self._obj, inverse=inverse, **kwargs)

        
    @bluebelt.core.decorators.class_methods
    class datetime():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
    
        # bluebelt.datetime.dt
        def year(self, column=None, **kwargs):
            return bluebelt.datetime.dt.year(self._obj[column], **kwargs)

        def quarter(self, column=None, **kwargs):
            return bluebelt.datetime.dt.quarter(self._obj[column], **kwargs)

        def month(self, column=None, **kwargs):
            return bluebelt.datetime.dt.month(self._obj[column], **kwargs)

        def day(self, column=None, **kwargs):
            return bluebelt.datetime.dt.day(self._obj[column], **kwargs)

        def weekday(self, column=None, **kwargs):
            return bluebelt.datetime.dt.weekday(self._obj[column], **kwargs)
        
        def weekday_name(self, column=None, **kwargs):
            return bluebelt.datetime.dt.weekday_name(self._obj[column], **kwargs)

        def is_holiday(self, column=None, **kwargs):
            return bluebelt.datetime.dt.is_holiday(self._obj[column], **kwargs)

        def is_weekend(self, column=None, **kwargs):
            return bluebelt.datetime.dt.is_weekend(self._obj[column], **kwargs)
        
        def iso_year(self, column=None, **kwargs):
            return bluebelt.datetime.dt.iso_year(self._obj[column], **kwargs)

        def iso_week(self, column=None, **kwargs):
            return bluebelt.datetime.dt.iso_week(self._obj[column], **kwargs)
        
        def week(self, column=None, **kwargs):
            return bluebelt.datetime.dt.week(self._obj[column], **kwargs)
        
        def date_from_weeknumber(self, year=None, week=None, **kwargs):
            return bluebelt.datetime.dt.date_from_weeknumber(self._obj, year=year, week=week, **kwargs)

        def add(self, column=None, prefix='_', **kwargs):
            self._obj.loc[:,f'{prefix}year'] = bluebelt.datetime.dt.year(self._obj[column])
            self._obj.loc[:,f'{prefix}quarter'] = bluebelt.datetime.dt.quarter(self._obj[column])
            self._obj.loc[:,f'{prefix}month'] = bluebelt.datetime.dt.month(self._obj[column])
            self._obj.loc[:,f'{prefix}day'] = bluebelt.datetime.dt.day(self._obj[column])
            self._obj.loc[:,f'{prefix}weekday'] = bluebelt.datetime.dt.weekday(self._obj[column])
            self._obj.loc[:,f'{prefix}day_name'] = bluebelt.datetime.dt.day_name(self._obj[column])
            self._obj.loc[:,f'{prefix}is_holiday'] = bluebelt.datetime.dt.is_holiday(self._obj[column])
            self._obj.loc[:,f'{prefix}is_weekend'] = bluebelt.datetime.dt.is_weekend(self._obj[column])
            self._obj.loc[:,f'{prefix}iso_year'] = bluebelt.datetime.dt.iso_year(self._obj[column])
            self._obj.loc[:,f'{prefix}iso_week'] = bluebelt.datetime.dt.iso_week(self._obj[column])
            return self._obj

    @bluebelt.core.decorators.class_methods
    class std():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        def std_within(self, axis=0, columns=None, how=None, observations=2, **kwargs):
            """
            Calculate the standard deviation within subgroups.
                arguments
                frame: pandas.DataFrame or pandas.Series
                columns: list
                    a list of column names
                    if not provided all columns will be treated as subgroups
                    if frame is a Pandas Series only the first column will be used
                    default value: None
                how: str
                    the method of estimating std_within
                    options when subgroup size == 1 (default when frame is a Pandas Series):
                    - 'amr' or 'average_moving_range'
                        optional argument: observations=2
                    - 'mmr' or 'median_moving_range'
                        optional argument: observations=2
                    - 'mssd' or 'sqrt_of_mssd'
                    default value: 'amr'

                    options when subgroup size > 1:
                    - 'pooled_std'
                    - 'r_bar'
                    - 's_bar'
                    default value: 'pooled_std'
                    
                observations: int
                    used for average moving range and median moving range estimation
                    default value: 2
            """
            # frame, axis=0, columns=None, how=None, observations=2, 
            return bluebelt.statistics.std.StdWithin(self._obj, axis=axis, columns=columns, how=how, observations=observations, **kwargs)
        
    @bluebelt.core.decorators.class_methods
    class patterns():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # patterns
        def anomalies(self, values=None, pattern=None):
            return bluebelt.analysis.patterns.anomalies(self._obj, values=values, pattern=pattern)

    @bluebelt.core.decorators.class_methods
    class statistics():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def correlation(self, columns=None, confidence=0.95, **kwargs):
            return bluebelt.statistics.basic_statistics.Correlation(self._obj, columns=columns, confidence=confidence)
     
    @bluebelt.core.decorators.class_methods
    class test():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def anderson_darling(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.AndersonDarling(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def normal_distribution(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.NormalDistribution(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def dagostino_pearson(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.DAgostinoPearson(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        # index
        @property
        def index(self):
            return bluebelt.statistics.hypothesis_testing.index()

        # hypothesis testing
        
        def equal_means(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.EqualMeans(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def two_sample_t(self, columns=None, related=False, confidence=0.95, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.TwoSampleT(self._obj, columns=columns, related=related, confidence=confidence, alpha=alpha, **kwargs)
        
        def wilcoxon(self, popmean=None, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Wilcoxon(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def anova(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Anova(self._obj, columns=columns, alpha=alpha, **kwargs)

        def kruskal(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.KruskalWallis(self._obj, columns=columns, alpha=alpha, **kwargs)

        kruskal_wallis = kruskal

        def mann_whitney(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.MannWhitney(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def levene(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Levene(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def bartlett(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Bartlett(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def f_test(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.FTest(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def equal_variance(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.EqualVariance(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        equal_var = equal_variance




import bluebelt.helpers.defaults as defaults

# boxplot | pyplot.boxplot
boxplot = {
    'boxes': {
        'hatch': '////',
        },
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


# matplotlib defaults
# Author: Arvid Hahné

from cycler import cycler
import matplotlib.pyplot as plt

# font sizes
xsmall = 10
small = 14
medium = 18
large = 21

# rgb colors to floats for matplotlib and plotly
white = tuple(ti/255 for ti in (255, 255, 255))
light_blue = tuple(ti/255 for ti in (63, 141, 197))
blue = tuple(ti/255 for ti in (37, 111, 179))
dark_blue = tuple(ti/255 for ti in (0, 17, 70))
yellow = tuple(ti/255 for ti in (252, 190, 31))
orange = tuple(ti/255 for ti in (238, 92, 43))
red = tuple(ti/255 for ti in (181, 17, 51))
green = tuple(ti/255 for ti in (13, 112, 52))
grey = tuple(ti/255 for ti in (140, 148, 168))
light_grey = tuple(ti/255 for ti in (181, 198, 207))
dark_grey = tuple(ti/255 for ti in (63, 73, 77))
black = tuple(ti/255 for ti in (14, 18, 33))

# plot styles
style = {
    'histogram': {
        'basic': {
            'fill': False,
            'linewidth': 1,
            'hatch': '//',
        },
        'color': {
            'linewidth': 1,
        }
    }
}

### LINES
# See http://matplotlib.org/api/artist_api.html#module-matplotlib.lines for more

plt.rc('lines', linewidth=1) # line width in points
plt.rc('lines', linestyle='-')# solid line
#plt.rc('lines', color='C0')# has no affect on plot(); see axes.prop_cycle
#plt.rc('lines', marker='o')# the default marker
plt.rc('lines', markeredgewidth=0)# the line width around the marker symbol
plt.rc('lines', markersize=5)# markersize, in points
plt.rc('lines', markerfacecolor=black)# marker face color, in points
#plt.rc('lines', dash_joinstyle='miter')# miter|round|bevel
plt.rc('lines', dash_capstyle='round')# butt|round|projecting
#plt.rc('lines', solid_joinstyle='miter')# miter|round|bevel
plt.rc('lines', solid_capstyle='round')# butt|round|projecting
#plt.rc('lines', antialiased='True')# render lines in antialiased (no jaggies)
# The three standard dash patterns.  These are scaled by the linewidth.
plt.rc('lines', dashed_pattern=(10, 5))
plt.rc('lines', dashdot_pattern=(10, 3, 1, 3,))
plt.rc('lines', dotted_pattern=(1, 3))
#plt.rc('lines', scale_dashes=True
#plt.rc('markers', fillstyle:='full')# full|left|right|bottom|top|none

### PATCHES
# http://matplotlib.org/api/artist_api.html#module-matplotlib.patches

plt.rc('patch', linewidth=0.5)# edge width in points.
plt.rc('patch', facecolor='none')
plt.rc('patch', edgecolor=black)# if forced, or patch is not filled
plt.rc('patch', force_edgecolor='True')# True to always use edgecolor
#plt.rc('patch', antialiased='True')# render patches in antialiased (no jaggies)

### HATCHES
plt.rc('hatch', color=light_grey)
plt.rc('hatch', linewidth=0.5)

### Boxplot
plt.rc('boxplot', notch=False)
plt.rc('boxplot', vertical=True)
plt.rc('boxplot', whiskers=1)
plt.rc('boxplot', bootstrap=None)
plt.rc('boxplot', patchartist=True)
plt.rc('boxplot', showmeans=True)
plt.rc('boxplot', showcaps=True)
plt.rc('boxplot', showbox=True)
plt.rc('boxplot', showfliers=True)
plt.rc('boxplot', meanline=True)

plt.rc('boxplot.flierprops', color=red)
plt.rc('boxplot.flierprops', marker='+')
#plt.rc('boxplot.flierprops', markerfacecolor=white)
plt.rc('boxplot.flierprops', markeredgecolor=red)
plt.rc('boxplot.flierprops', markersize=7)
plt.rc('boxplot.flierprops', linestyle='none')
plt.rc('boxplot.flierprops', linewidth=1)

plt.rc('boxplot.boxprops', color=black)
plt.rc('boxplot.boxprops', linewidth=1)
plt.rc('boxplot.boxprops', linestyle='-')

#vertical lines Q1 & Q4
plt.rc('boxplot.whiskerprops', color=black)
plt.rc('boxplot.whiskerprops', linewidth=1)
plt.rc('boxplot.whiskerprops', linestyle='-')

#horizontal lines min & max
plt.rc('boxplot.capprops', color=black)
plt.rc('boxplot.capprops', linewidth=1)
plt.rc('boxplot.capprops', linestyle='-')

plt.rc('boxplot.medianprops', color=black)
plt.rc('boxplot.medianprops', linewidth=1)
plt.rc('boxplot.medianprops', linestyle='-')

plt.rc('boxplot.meanprops', color=blue)
plt.rc('boxplot.meanprops', marker=None)
#plt.rc('boxplot.meanprops', markerfacecolor=blue)
#plt.rc('boxplot.meanprops', markeredgecolor=white)
#plt.rc('boxplot.meanprops', markersize=7)
plt.rc('boxplot.meanprops', linestyle='solid')
plt.rc('boxplot.meanprops', linewidth=1)

### FONT
plt.rc('font', family='sans-serif') #serif', 'sans-serif', 'cursive', 'fantasy', and 'monospace'
plt.rc('font', style='normal') #normal (or roman), italic or oblique
plt.rc('font', variant='normal') #normal or small-caps
plt.rc('font', weight='normal') #normal, bold, bolder, lighter, 100, 200, 300, ..., 900
plt.rc('font', stretch='normal') #ultra-condensed, extra-condensed, condensed, semi-condensed, normal, semi-expanded, expanded, extra-expanded, ultra-expanded, wider, and narrower
plt.rc('font', size=small)
plt.rc('font', serif='Times')
#plt.rc('font', sans-serif='Tahoma')
plt.rc('font',**{'sans-serif':['Tahoma']})
#plt.rc('font', cursive='Zapf-Chancery')
#plt.rc('font', fantasy='Comic')
#plt.rc('font', monospace='Courier')

### TEXT
# http://matplotlib.org/api/artist_api.html#module-matplotlib.text for more
plt.rc('text', color=black)

### LaTeX customizations. See http://wiki.scipy.org/Cookbook/Matplotlib/UsingTex
#plt.rc('text', usetex='False')# use latex for all text handling.
#plt.rc('text', latex.unicode='False')# use "ucs" and "inputenc" LaTeX packages for handling unicode strings.
#plt.rc('text', latex.preamble='')# IMPROPER USE OF THIS FEATURE WILL LEAD TO LATEX FAILURES
#plt.rc('text', dvipnghack='None')# When dvipng doesn't handle alpha channel properly. True, False or None (=Guess)
#plt.rc('text', hinting='auto')# May be one of the following: 'none', 'auto', 'native', 'either'
#plt.rc('text', hinting_factor=8)# Specifies the amount of softness for hinting in the horizontal direction.
#plt.rc('text', antialiased='True')# If True (default), the text will be antialiased. This only affects the Agg backend.

# The following settings allow you to select the fonts in math mode. These settings are only used if mathtext.fontset is 'custom'. Note that this "custom" mode is unsupported and may go away in the future.
#plt.rc('mathtext', cal='cursive')
#plt.rc('mathtext', rm='serif')
#plt.rc('mathtext', tt='monospace')
#plt.rc('mathtext', it='serif:italic')
#plt.rc('mathtext', bf='serif:bold')
#plt.rc('mathtext', sf='sans')
#plt.rc('mathtext', fontset='dejavusans')# Should be 'dejavusans' (default), 'dejavuserif', 'cm' (Computer Modern), 'stix', 'stixsans' or 'custom'
#plt.rc('mathtext', fallback_to_cm=True)# When True, use symbols from the Computer Modern fonts when a symbol can not be found in one of the custom math fonts.
#plt.rc('mathtext', default='it')# The default font to use for math.

### AXES
# http://matplotlib.org/api/axes_api.html#module-matplotlib.axes
plt.rc('axes', facecolor=white)# axes background color
plt.rc('axes', edgecolor=black)# axes edge color
plt.rc('axes', linewidth=1)# edge linewidth
plt.rc('axes', grid=True)# display grid or not
plt.rc('axes.grid', axis='both')# which axis (x, y, both)
plt.rc('axes.grid', which='major')# major, minor or both

plt.rc('axes', titlesize=medium)# fontsize of the axes title
#plt.rc('axes', titlelocation='left')# title location (left, right, center)
#plt.rc('axes', titley=1.0)# pad between title and axis in points
plt.rc('axes', titlepad=10)# pad between title and axis in points
plt.rc('axes', labelsize=small)# fontsize of the x any y labels
#plt.rc('axes', labelpad=4.0)# space between label and axis
#plt.rc('axes', labelweight='normal')# weight of the x and y labels
plt.rc('axes', labelcolor=black)
#plt.rc('axes', axisbelow='line')# draw axis gridlines and ticks below patches (True); above patches but below lines ('line'); or above all (False)
#plt.rc('axes.formatter', limits='-7,')# use scientific notation if log10 of the axis range is smaller than the first or larger than the second
#plt.rc('axes.formatter', use_locale='False')# When True, format tick labels according to the user's locale. For example, use ',' as a decimal separator in the fr_FR locale.
#plt.rc('axes.formatter', use_mathtext='False')# When True, use mathtext for scientific notation.
#plt.rc('axes.formatter', useoffset='True')# If True, the tick label formatter will default to labeling ticks relative to an offset when the data range is small compared to the minimum absolute value of the data.
#plt.rc('axes.formatter', offset_threshold=4)# When useoffset is True, the offset will be used when it can remove at least this number of significant digits from tick labels.
plt.rc('axes.spines', left=True)  # display axis spines
plt.rc('axes.spines', bottom=True)
plt.rc('axes.spines', top=True)
plt.rc('axes.spines', right=True)
#plt.rc('axes', unicode_minus=True)# use unicode for the minus symbol rather than hyphen. See http://en.wikipedia.org/wiki/Plus_and_minus_signs#Character_codes
plt.rc('axes', prop_cycle=(cycler('color', [blue, light_blue, orange, yellow, green, dark_blue, red, grey]))) # color cycle for plot lines
plt.rc('axes', autolimit_mode='data')# How to scale axes limits to the data. Use "data" to use data limits, plus some margin. Use "round_numbers" move to the nearest "round" number
plt.rc('axes', xmargin='.05')# x margin. See `axes.Axes.margins`
plt.rc('axes', ymargin='.25')# y margin. See `axes.Axes.margins`
#plt.rc('polaraxes', grid='True')# display grid on polar axes
#plt.rc('axes3d', grid='True')# display grid on 3d axes

### DATES
# These control the default format strings used in AutoDateFormatter.
# Any valid format datetime format string can be used (see the python
# `datetime` for details).  For example using '%%x' will use the locale date representation
# '%%X' will use the locale time representation and '%%c' will use the full locale datetime
# representation.
# These values map to the scales:
#    {'year': 365, 'month': 30, 'day': 1, 'hour': 1/24, 'minute': 1 / (24 * 60)}
#plt.rc('date.autoformatter', year='%Y')
#plt.rc('date.autoformatter', month='%Y-%m')
#plt.rc('date.autoformatter', day='%Y-%m-%d')
#plt.rc('date.autoformatter', hour='%m-%d %H')
#plt.rc('date.autoformatter', minute='%d %H:%M')
#plt.rc('date.autoformatter', second='%H:%M:%S')
#plt.rc('date.autoformatter', microsecond='%M:%S.%f')

### TICKS
# see http://matplotlib.org/api/axis_api.html#matplotlib.axis.Tick
plt.rc('xtick', top=True)# draw ticks on the top side
plt.rc('xtick', bottom=True)# draw ticks on the bottom side
plt.rc('xtick', color=black)# color of the tick labels
plt.rc('xtick', labelsize=small)# fontsize of the tick labels
plt.rc('xtick', direction='inout')# direction: in, out, or inout

plt.rc('xtick.major', size=5)# major tick size in points
plt.rc('xtick.major', width=1)# major tick width in points
#plt.rc('xtick.major', pad=3.5)# distance to major tick label in points
plt.rc('xtick.major', top=True)# draw x axis top major ticks
plt.rc('xtick.major', bottom=True)# draw x axis bottom major ticks

plt.rc('xtick.minor', visible=False)# visibility of minor ticks on x-axis
#plt.rc('xtick.minor', size=0)# minor tick size in points
#plt.rc('xtick.minor', width=0.6)# minor tick width in points
#plt.rc('xtick.minor', pad=3.4)# distance to the minor tick label in points
#plt.rc('xtick.minor', top=False)# draw x axis top minor ticks
#plt.rc('xtick.minor', bottom=False)# draw x axis bottom minor ticks

plt.rc('ytick', left=True)# draw ticks on the left side
plt.rc('ytick', right=True)# draw ticks on the right side
plt.rc('ytick', color=black)# color of the tick labels
plt.rc('ytick', labelsize=small)# fontsize of the tick labels
plt.rc('ytick', direction='inout')# direction: in, out, or inout

#plt.rc('ytick.major', size=5)# major tick size in points
#plt.rc('ytick.major', width=0.8)# major tick width in points
#plt.rc('ytick.major', pad=3.5)# distance to major tick label in points
#plt.rc('ytick.major', left=True)# draw y axis left major ticks
#plt.rc('ytick.major', right=False)# draw y axis right major ticks

plt.rc('ytick.minor', visible=False)# visibility of minor ticks on y-axis
#plt.rc('ytick.minor', size=0)# minor tick size in points
#plt.rc('ytick.minor', width=0.6)# minor tick width in points
#plt.rc('ytick.minor', pad=3.4)# distance to the minor tick label in points
#plt.rc('ytick.minor', left=False)# draw y axis left minor ticks
#plt.rc('ytick.minor', right=False)# draw y axis right minor ticks

### GRIDS
plt.rc('grid', color=light_grey)# grid color
plt.rc('grid', linestyle='dotted')# solid
plt.rc('grid', linewidth=0.5)# in points
plt.rc('grid', alpha=0.5)# transparency, between 0.0 and 1.0

### Legend
plt.rc('legend', loc='best')
#plt.rc('legend', frameon=False)# if True, draw the legend on a background patch
#plt.rc('legend', framealpha=0.8)# legend patch transparency
#plt.rc('legend', facecolor='inherit')# inherit from axes.facecolor; or color spec
#plt.rc('legend', edgecolor=0.8)# background patch boundary color
#plt.rc('legend', fancybox=False)# if True, use a rounded box for the
                                 # legend background, else a rectangle
#plt.rc('legend', shadow='False')# if True, give background a shadow effect
#plt.rc('legend', numpoints=1)# the number of marker points in the legend line
#plt.rc('legend', scatterpoints=1)# number of scatter points
#plt.rc('legend', markerscale=1.0)# the relative size of legend markers vs. original
plt.rc('legend', fontsize=xsmall)
# Dimensions as fraction of fontsize:
#plt.rc('legend', borderpad=0.4)# border whitespace
#plt.rc('legend', labelspacing=0.5)# the vertical space between the legend entries
#plt.rc('legend', handlelength=2.0)# the length of the legend lines
#plt.rc('legend', handleheight=0.7)# the height of the legend handle
#plt.rc('legend', handletextpad=0.8)# the space between the legend line and legend text
#plt.rc('legend', borderaxespad=0.5)# the border between the axes and legend edge
#plt.rc('legend', columnspacing=2.0)# column separation

### FIGURE
# See http://matplotlib.org/api/figure_api.html#matplotlib.figure.Figure
plt.rc('figure', titlesize=large)# size of the figure title (Figure.suptitle())
#plt.rc('figure', titleweight='normal')# weight of the figure title
plt.rc('figure', figsize=(10,7))# figure size in inches
#plt.rc('figure', dpi=100)# figure dots per inch
plt.rc('figure', facecolor=white)# figure facecolor; 0.75 is scalar gray
plt.rc('figure', edgecolor=white)# figure edgecolor
#plt.rc('figure', autolayout='False')# When True, automatically adjust subplot
                            # parameters to make the plot fit the figure
#plt.rc('figure', max_open_warning=20)# The maximum number of figures to open through
                               # the pyplot interface before emitting a warning.
                               # If less than one this feature is disabled.
# The figure subplot parameters.  All dimensions are a fraction of the
#plt.rc('figure', subplot.left=0.125)# the left side of the subplots of the figure
#plt.rc('figure', subplot.right=0.9)# the right side of the subplots of the figure
#plt.rc('figure', subplot.bottom=0.11)# the bottom of the subplots of the figure
#plt.rc('figure', subplot.top=0.88)# the top of the subplots of the figure
#plt.rc('figure', subplot.wspace=0.2)# the amount of width reserved for blank space between subplots,
                                 # expressed as a fraction of the average axis width
#plt.rc('figure', subplot.hspace=0.2)# the amount of height reserved for white space between subplots,
                                 # expressed as a fraction of the average axis height
### IMAGES
#plt.rc('image', aspect='equal')# equal | auto | a number
#plt.rc('image', interpolation='nearest')# see help(imshow) for options
#plt.rc('image', cmap='viridis')# A colormap name, gray etc...
#plt.rc('image', lut=256)# the size of the colormap lookup table
#plt.rc('image', origin='upper')# lower | upper
#plt.rc('image', resample=
#plt.rc('image', composite_image='True')# When True, all the images on a set of axes are
                                  # combined into a single composite image before
                                  # saving a figure as a vector graphics file,
                                  # such as a PDF.
### CONTOUR PLOTS
#plt.rc('contour', negative_linestyle='dashed')# dashed | solid
#plt.rc('contour', corner_mask='True')# True | False | legacy

### ERRORBAR PLOTS
#plt.rc('errorbar', capsize=0)# length of end cap on error bars in pixels

### HISTOGRAM PLOTS
plt.rc('hist', bins='auto')# The default number of histogram bins.
                                  # If Numpy 1.11 or later is
                                  # installed, may also be `auto`
### SCATTER PLOTS
plt.rc('scatter', marker='o')# The default marker type for scatter plots.
### Agg rendering
### Warning: experimental, 2008/10/10
#plt.rc('agg', path.chunksize=0)# 0 to disable; values in the range
                                  # 10000 to 100000 can improve speed slightly
                                  # and prevent an Agg rendering failure
                                  # when plotting very large data sets,
                                  # especially if they are very gappy.
                                  # It may cause minor artifacts, though.
                                  # A value of 20000 is probably a good
                                  # starting point.
### SAVING FIGURES
#plt.rc('path', simplify='True')# When True, simplify paths by removing "invisible"
                        # points to reduce file size and increase rendering
                        # speed
#plt.rc('path', simplify_threshold=0.1)# The threshold of similarity below which
                                # vertices will be removed in the simplification
                                # process
#plt.rc('path', snap='True')# When True, rectilinear axis-aligned paths will be snapped to
                  # the nearest pixel when certain criteria are met.  When False,
                  # paths will never be snapped.
#plt.rc('path', sketch='None')# May be none, or a 3-tuple of the form (scale, length,
                    # randomness).
                    # *scale* is the amplitude of the wiggle
                    # perpendicular to the line (in pixels).  *length*
                    # is the length of the wiggle along the line (in
                    # pixels).  *randomness* is the factor by which
                    # the length is randomly scaled.
# the default savefig params can be different from the display params
# e.g., you may want a higher resolution, or to make the figure
# background white
#plt.rc('savefig', dpi='figure')# figure dots per inch or 'figure'
#plt.rc('savefig', facecolor='white')# figure facecolor when saving
#plt.rc('savefig', edgecolor='white')# figure edgecolor when saving
#plt.rc('savefig', format='png')# png, ps, pdf, svg
#plt.rc('savefig', bbox='standard')# 'tight' or 'standard'.
                                # 'tight' is incompatible with pipe-based animation
                                # backends but will workd with temporary file based ones:
                                # e.g. setting animation.writer to ffmpeg will not work,
                                # use ffmpeg_file instead
#plt.rc('savefig', pad_inches=0.1)# Padding to be used when bbox is set to 'tight'
#plt.rc('savefig', jpeg_quality:=95)# when a jpeg is saved, the default quality parameter.
#plt.rc('savefig', directory='~')# default directory in savefig dialog box,
                                # leave empty to always use current working directory
#plt.rc('savefig', transparent='False')# setting that controls whether figures are saved with a
                                # transparent background by default

# tk backend params
#plt.rc('tk', window_focus='False')# Maintain shell focus for TkAgg
# ps backend params
#plt.rc('ps', papersize='letter')# auto, letter, legal, ledger, A0-A10, B0-B10
#plt.rc('ps', useafm='False')# use of afm fonts, results in small files
#plt.rc('ps', usedistiller='False')# can be: None, ghostscript or xpdf
                                          # Experimental: may produce smaller files.
                                          # xpdf intended for production of publication quality files,
                                          # but requires ghostscript, xpdf and ps2eps
#plt.rc('ps', distiller.res=6000)# dpi
#plt.rc('ps', fonttype=3)# Output Type 3 (Type3) or Type 42 (TrueType)
# pdf backend params
#plt.rc('pdf', compression=6)# integer from 0 to 9
                       # 0 disables compression (good for debugging)
#plt.rc('pdf', fonttype=3)# Output Type 3 (Type3) or Type 42 (TrueType)
# svg backend params
#plt.rc('svg', image_inline='True')# write raster image data directly into the svg file
#plt.rc('svg', fonttype=''path'')# How to handle SVG fonts:
#    'none': Assume fonts are installed on the machine where the SVG will be viewed.
#    'path': Embed characters as paths -- supported by most SVG renderers
#    'svgfont': Embed characters as SVG fonts -- supported only by Chrome,
#               Opera and Safari
#plt.rc('svg', hashsalt='None')# if not None, use this string as hash salt
                               # instead of uuid4
# docstring params
#plt.rc('docstring', hardcopy=# set this when you want to generate hardcopy docstring
# Set the verbose flags.  This controls how much information
# matplotlib gives you at runtime and where it goes.  The verbosity
# levels are: silent, helpful, debug, debug-annoying.  Any level is
# inclusive of all the levels below it.  If your setting is "debug",
# you'll get all the debug and helpful messages.  When submitting
# problems to the mailing-list, please set verbose to "helpful" or "debug"
# and paste the output into your report.
#
# The "fileo" gives the destination for any calls to verbose.report.
# These objects can a filename, or a filehandle like sys.stdout.
#
# You can override the rc default verbosity from the command line by
# giving the flags --verbose-LEVEL where LEVEL is one of the legal
# levels, e.g., --verbose-helpful.
#
# You can access the verbose instance in your code
#   from matplotlib import verbose.
#plt.rc('verbose', level='silent')# one of silent, helpful, debug, debug-annoying
#plt.rc('verbose', fileo='sys.stdout')# a log filename, sys.stdout or sys.stderr
# Event keys to interact with figures/plots via keyboard.
# Customize these settings according to your needs.
# Leave the field(s) empty if you don't need a key-map. (i.e., fullscreen : '')
#plt.rc('keymap', fullscreen='f,')# toggling
#plt.rc('keymap', home='h,')# home or reset mnemonic
#plt.rc('keymap', back='left,')# forward / backward keys to enable
#plt.rc('keymap', forward='right,')#   left handed quick navigation
#plt.rc('keymap', pan='p')# pan mnemonic
#plt.rc('keymap', zoom='o')# zoom mnemonic
#plt.rc('keymap', save='s')# saving current figure
#plt.rc('keymap', quit='ctrl+w,')# close the current figure
#plt.rc('keymap', grid='g')# switching on/off a grid in current axes
#plt.rc('keymap', yscale='l')# toggle scaling of y-axes ('log'/'linear')
#plt.rc('keymap', xscale='L,')# toggle scaling of x-axes ('log'/'linear')
#plt.rc('keymap', all_axes='a')# enable all axes
# Control location of examples data files
#plt.rc('examples', directory='''')# directory to look in for custom installation
###ANIMATION settings
#plt.rc('animation', html=''none'')# How to display the animation as HTML in
                                   # the IPython notebook. 'html5' uses
                                   # HTML5 video tag.
#plt.rc('animation', writer='ffmpeg')# MovieWriter 'backend' to use
#plt.rc('animation', codec='h264')# Codec to use for writing movie
#plt.rc('animation', bitrate:='-1')# Controls size/quality tradeoff for movie.
                                   # -1 implies let utility auto-determine
#plt.rc('animation', frame_format:=''png'')# Controls frame format used by temp files
#plt.rc('animation', ffmpeg_path:=''ffmpeg'')# Path to ffmpeg binary. Without full path
                                   # $PATH is searched
#plt.rc('animation', ffmpeg_args:='''')# Additional arguments to pass to ffmpeg
#plt.rc('animation', avconv_path:=''avconv'')# Path to avconv binary. Without full path
                                   # $PATH is searched
#plt.rc('animation', avconv_args:='''')# Additional arguments to pass to avconv
#plt.rc('animation', mencoder_path:=
                                   # Path to mencoder binary. Without full path
                                   # $PATH is searched
#plt.rc('animation', mencoder_args:='''')# Additional arguments to pass to mencoder
#plt.rc('animation', convert_path:=''convert'')# Path to ImageMagick's convert binary.
                                   # On Windows use the full path since convert
                                   # is also the name of a system tool.



import pandas as pd
import numpy as np
import scipy.stats as stats


import bluebelt.datetime.dt

import bluebelt.statistics.std

import bluebelt.analysis.ci
import bluebelt.statistics.hypothesis_testing
import bluebelt.analysis.patterns
import bluebelt.analysis.datetime
import bluebelt.analysis.graphs
import bluebelt.analysis.ppa

import bluebelt.core.decorators

@bluebelt.core.decorators.class_methods
@pd.api.extensions.register_series_accessor('blue')
class SeriesToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.std = self.std(self._obj)
        self.patterns = self.patterns(self._obj)
        self.datetime = self.datetime(self._obj)
        self.data = self.data(self._obj)
        self.graphs = self.graphs(self._obj)
        self.statistics = self.statistics(self._obj)
        self.test = self.test(self._obj)
        self.ppa = self.ppa(self._obj)
    
    @bluebelt.core.decorators.class_methods
    class std():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        # standard deviation
        def std_within(self, how=None, observations=2, **kwargs):
            """
            Calculate the standard deviation within subgroups.
                arguments
                frame: pandas.DataFrame or pandas.Series
                columns: list
                    a list of column names
                    if not provided all columns will be treated as subgroups
                    if frame is a Pandas Series only the first column will be used
                    default value: None
                how: str
                    the method of estimating std_within
                    options when subgroup size == 1 (default when frame is a Pandas Series):
                    - 'amr' or 'average_moving_range'
                        optional argument: observations=2
                    - 'mmr' or 'median_moving_range'
                        optional argument: observations=2
                    - 'mssd' or 'sqrt_of_mssd'
                    default value: 'amr'

                    options when subgroup size > 1:
                    - 'pooled_std'
                    - 'r_bar'
                    - 's_bar'
                    default value: 'pooled_std'

                observations: int
                    used for average moving range and median moving range estimation
                    default value: 2
            """
            return bluebelt.statistics.std.StdWithin(self._obj, how=how, observations=observations, **kwargs)        

    @bluebelt.core.decorators.class_methods
    class patterns():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        # patterns
        def polynomial(self, **kwargs):
            """
            Find the polynomial of a series and project a bandwidth
                arguments
                series: pandas.Series
                shape: int or tuple
                    when an int is provided the polynomial is provided as n-th degree polynomial
                    when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
                    default value: (0, 6)
                validation: string
                    validation type for shape tuple
                    p_val: test for normal distribution of the residuals
                    rsq: check for improvement of the rsq value
                    default value: p_val
                threshold: float
                    the threshold for normal distribution test or rsq improvement
                    default value: 0.05
                confidence: float
                    the bound confidence
                    default value: 0.8
                outlier_sigma: float
                    outliers are datapoints outside the outlier_sigma fraction
                    default value: 2
                adjust: boolean
                    adjust polynomial for outliers
                    default value: True
            """
            return bluebelt.analysis.patterns.Polynomial(self._obj, **kwargs)

        def periodical(self, **kwargs):
            """
            Find the periodical pattern of a series and project a bandwidth
                series: pandas.Series
                rule: period representation used for resampling the series
                    default value: "1W"
                how: define how the period must be evaluated
                    options are "mean", "min", "max" and "std"
                    default value: "mean"
                resolution: define the resolution of the pattern
                    the pattern is rounded to fit the resolution
                    default value: None
                confidence: float
                    the bandwidth confidence
                    default value: 0.8
                outlier_sigma: float
                    outliers are datapoints outside the outlier_sigma fraction
                    default value: 2    
            """
            return bluebelt.analysis.patterns.Periodical(self._obj, **kwargs)
    
    @bluebelt.core.decorators.class_methods
    class datetime():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # week day
        def week_day(self, **kwargs):
            """
            Compare the distribution of data between week days
                arguments
                series: pandas.Series
                    the Series must have a pandas.DatetimeIndex
                
                properties
                .series
                    the transformed pandas.Series with week day index
                .data
                    the data
                .equal_means
                    the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
                .equal_variance
                    the result of bluebelt.statistics.hypothesis_testing.EqualVariance().passed

                methods
                .plot()
                    plot the results as a boxplot
            """
            return bluebelt.analysis.datetime.WeekDay(self._obj, **kwargs)

        weekday = week_day

        # day of the month
        def month_day(self, **kwargs):
            """
            Compare the distribution of data between month days
                arguments
                series: pandas.Series
                    the Series must have a pandas.DatetimeIndex
                
                properties
                .series
                    the transformed pandas.Series with month day index
                .data
                    the data
                .equal_means
                    the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
                .equal_variance
                    the result of bluebelt.statistics.hypothesis_testing.EqualVariance().passed

                methods
                .plot()
                    plot the results as a boxplot
            """
            return bluebelt.analysis.datetime.MonthDay(self._obj, **kwargs)

        day = month_day

        # week of the year
        def week(self, **kwargs):
            """
            Compare the distribution of data between weeks
                arguments
                series: pandas.Series
                    the Series must have a pandas.DatetimeIndex
                
                properties
                .series
                    the transformed pandas.Series with week number index
                .data
                    the data
                .equal_means
                    the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
                .equal_variance
                    the result of bluebelt.statistics.hypothesis_testing.EqualVariance().passed

                methods
                .plot()
                    plot the results as a boxplot
            """
            return bluebelt.analysis.datetime.Week(self._obj, **kwargs)

        # month of the year
        def month(self, **kwargs):
            """
            Compare the distribution of data between months
                arguments
                series: pandas.Series
                    the Series must have a pandas.DatetimeIndex
                
                properties
                .series
                    the transformed pandas.Series with month index
                .data
                    the data
                .equal_means
                    the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
                .equal_variance
                    the result of bluebelt.statistics.hypothesis_testing.EqualVariance().passed

                methods
                .plot()
                    plot the results as a boxplot
            """
            return bluebelt.analysis.datetime.Month(self._obj, **kwargs)

        # year
        def year(self, **kwargs):
            """
            Compare the distribution of data between years
                arguments
                series: pandas.Series
                    the Series must have a pandas.DatetimeIndex
                
                properties
                .series
                    the transformed pandas.Series with year index
                .data
                    the data
                .equal_means
                    the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
                .equal_variance
                    the result of bluebelt.statistics.hypothesis_testing.EqualVariance().passed

                methods
                .plot()
                    plot the results as a boxplot
            """
            return bluebelt.analysis.datetime.Year(self._obj, **kwargs) 

    @bluebelt.core.decorators.class_methods
    class data():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        def resample(self, **kwargs):
            """
            Resample as series and apply a specific function.
                arguments
                series: pandas.Series
                rule: str
                    the resampling rule
                any pandas.Series.resample argunment

                Apply one of the following functions:
                    .sum()
                    .mean()
                    .min()
                    .max()
                    .std()
                    .value_range()
                    .count()
                    .subsize_count()
                    .subseries_count()

                e.g. series.blue.data.resample(rule="1W").sum()
            """
            return bluebelt.data.resolution.Resample(self._obj, **kwargs)
        
        def group_index(self, **kwargs):
            """
            Group a series by DateTime index and apply a specific function.
                arguments
                series: pandas.Series
                index: str
                    a string with date-time keywords that can be parsed to group the index
                    keywords:
                        weekday_name -> %a  # abbreviated weekday name
                        week_day -> %u      # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                        weekday -> %u       # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                        iso_week -> %V      # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                        isoweek -> %V       # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                        monthname -> %b     # abbreviated month name
                        month_name -> %b    # abbreviated month name
                        monthday -> %d      # day of the month (01 to 31)
                        month_day -> %d     # day of the month (01 to 31)
                        isoyear -> %G       # 4-digit year corresponding to the ISO week number (see %V).
                        iso_year -> %G      # 4-digit year corresponding to the ISO week number (see %V).
                        yearday -> %j       # day of the year (001 to 366)
                        year_day -> %j      # day of the year (001 to 366)
                        year -> %Y          # year including the century
                        month -> %m         # month (01 to 12)
                        week -> %W          # week number of the current year, starting with the first Monday as the first day of the first week
                        day -> %d           # day of the month (01 to 31)
                        hour -> %H          # hour, using a 24-hour clock (00 to 23)
                        minute -> %M        # minute
                        second -> %S        # second
                        time -> %T          # current time, equal to %H:%M:%S
                    
                    default value "iso_year-iso_week"

                    e.g. "iso_year-Wiso_week" will parse to "%G-W%V" which will print "2021-W01" for the first week of 2021

                Apply one of the following functions:
                    .sum()
                    .mean()
                    .min()
                    .max()
                    .std()
                    .value_range()
                    .count()
                    .subsize_count()
                    .subseries_count()

                e.g. series.blue.data.group_index(index="iso_year-iso_week").sum()
                
            """
            return bluebelt.data.resolution.GroupByDatetimeIndex(self._obj, **kwargs)
        
    @bluebelt.core.decorators.class_methods
    class graphs():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        # rolling std
        def rolling_std(self, **kwargs):
            return bluebelt.analysis.graphs.RollingStd(self._obj, **kwargs)

    @bluebelt.core.decorators.class_methods
    class statistics():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
    
    @bluebelt.core.decorators.class_methods
    class test():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # index
        @property
        def index(self):
            return bluebelt.statistics.hypothesis_testing.index()

        # hypothesis testing
        def normal_distribution(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.NormalDistribution(self._obj, alpha=alpha)
        
        def dagostino_pearson(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.DAgostinoPearson(self._obj, alpha=alpha)
        
        def anderson_darling(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.AndersonDarling(self._obj, alpha=alpha)
                
        def one_sample_t(self, popmean=None, confidence=0.95, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.OneSampleT(self._obj, popmean=popmean, confidence=confidence, alpha=alpha, **kwargs)

        def wilcoxon(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Wilcoxon(self._obj, alpha=alpha, **kwargs)
        
    @bluebelt.core.decorators.class_methods
    class ppa():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def summary(self, **kwargs):
            return bluebelt.analysis.ppa.Summary(self._obj, **kwargs)
        
        def control_chart(self, **kwargs):
            return bluebelt.analysis.ppa.ControlChart(self._obj, **kwargs)
            
        def run_chart(self, alpha=0.05, **kwargs):
            return bluebelt.analysis.ppa.RunChart(self._obj, alpha=alpha, **kwargs)

        def process_capability(self, **kwargs):
            '''
            Calculate and display the process capability
                
                arguments
                target: float
                    target value for the process
                    default value: None
                usl: float
                    upper specification limit (usl and ub cannot be specified both)
                    default value: None
                ub: float
                    upper bound (usl and ub cannot be specified both)
                    default value: None
                lsl: float
                    lower specification limit
                    default value: None
                lb: float
                    lower bound (lsl and lb cannot be specified both)
                    default value: None
                tolerance: float
                    sigma tolerance of the process
                    default value: 6.0

                methods
                .md()
                    show the process capability as markdown
                .df()
                    show the process capability in a pandas.DataFrame
                .plot()
                    plot the process capability
                    
                properties
                @ observed performance
                .observed_lt_lsl
                .observed_gt_usl
                .observed_performance

                @ expected performance
                .expected_lt_lsl_within
                .expected_gt_usl_within
                .expected_performance_within

                .expected_lt_lsl_overall
                .expected_gt_usl_overall
                .expected_performance_overall
                
                @ within capability
                .cp
                .cpl
                .cpu
                .cpk
                .ccpk
                
                @ overall capability
                .pp
                .ppl
                .ppu
                .ppk
                .cpm
            '''
            return bluebelt.analysis.ppa.ProcessCapability(self._obj, **kwargs)
        
        


import pandas as pd
import scipy.stats as stats

import matplotlib.pyplot as plt

from bluebelt.core import graphs
import bluebelt.helpers.mpl_format as mpl_format

import bluebelt.styles

class RollingStd():
    
    """
    Show the rolling standard deviation of a pandas series
    series: pandas.Series
    window: int
        the window size
        default value: 7
    center: boolean
        center the labels
        default value: True
    accepts all arguments for pd.Series.Rolling
    """

    def __init__(self, series, window=7, center=True, **kwargs):
        
        self.series = series
        self.confidence = kwargs.pop('confidence', None)
        self.window = window
        self.center = center
        self.win_type = kwargs.get('win_type', None)
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        if self.confidence:
            sigma_level = stats.norm.ppf(1-(1-self.confidence)/2) * 2
        else:
            sigma_level = 1
        self.rolling = self.series.dropna().rename(f'{self.series.name} rolling std').rolling(window=self.window, center=self.center, **kwargs).std() * sigma_level
        

    def plot(self, **kwargs):
        
        title = kwargs.pop('title', f'{self.series.name} rolling standard deviation')
        style = kwargs.pop('style', bluebelt.styles.paper)
        
        path = kwargs.pop('path', None)        
        
        # prepare figure
        fig, ax = plt.subplots(**kwargs)
        
        # area plot
        ax.stackplot(self.series.index, self.series.values, **style.graphs.area.stackplot)
        ax.plot(self.series, **style.graphs.area.plot)
        
        # labels
        ax.set_title(title, **style.graphs.area.title)
        
        # set x axis locator
        mpl_format.axisformat(ax, self.series)
        ax.set_ylim(0,)

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig


import numpy as np
import pandas as pd
from datetime import datetime, timedelta, date
from bluebelt.datetime import holidays

# series

def add_datetime_metadata(series):
    frame = pd.DataFrame(data={'datetime': series.index.values, 'values': series.values})
    frame['year'] = frame['datetime'].dt.year
    frame['quarter'] = frame['datetime'].dt.quarter
    frame['month'] = frame['datetime'].dt.month
    frame['day'] = frame['datetime'].dt.day
    frame['weekday'] = frame['datetime'].dt.weekday
    frame['day_name'] = frame['datetime'].dt.day_name()
    frame['iso_year'] = frame['datetime'].dt.isocalendar().year
    frame['iso_week'] = frame['datetime'].dt.isocalendar().week

    return frame
    

def year(series, **kwargs):
    return series.dt.year

def quarter(series, **kwargs):
    return series.dt.quarter

def month(series, **kwargs):
    return series.dt.month

def day(series, **kwargs):
    return series.dt.day

def weekday(series, **kwargs):
    return series.dt.weekday

def day_name(series, **kwargs):
    return series.dt.day_name()

def is_holiday(series, **kwargs):
    return series.apply(lambda x: holidays.holidays(x))

def is_weekend(series, **kwargs):
    weekend_days = kwargs.get('weekend_days', [5, 6])
    return series.apply(lambda x: pd.to_datetime(x).date().weekday() in weekend_days)

def iso_year(series, **kwargs):
    return series.dt.isocalendar().year

def iso_week(series, **kwargs):
    return series.dt.isocalendar().week

def week(series, **kwargs):
    return series.dt.isocalendar().week

# dataframe
def date_from_weeknumber(df,
                         year=None,
                         week=None,
                         day=0,
                         ):


    # data checks
    if isinstance(year, (list, pd.Series, np.ndarray, int, float)) and isinstance(week, (list, pd.Series, np.ndarray, int, float)):
        year = pd.Series(year)
        week = pd.Series(week)
        if year.shape[0] == week.shape[0]:
            df = pd.DataFrame({'year': year, 'week': week})
        else:
            raise ValueError("'year' and 'week' need to have equal lengths")
    elif isinstance(df, pd.DataFrame) and isinstance(year, str) and isinstance(week, str):
        df = pd.DataFrame({'year': df[year], 'week': df[week]})
    else:
        raise ValueError("arguments not in the right format")

    def weeknum(row, year=None, week=None, day=0):
        d = pd.to_datetime(f'4 jan {row[year]}')  # The Jan 4th must be in week 1 according to ISO
        result =  d + pd.Timedelta(weeks=(row[week])-1, days=-(d.weekday()) + day)
        return result
    
    return df.apply(lambda x: weeknum(x, year=year, week=week, day=day), axis=1)



import re
import math

from scipy.stats import distributions

def get_nice_list(the_list, max_len=3):
    if len(the_list) == 1:
        return the_list[0]
    elif len(the_list) <= max_len:
        return ", ".join(the_list[:-1])+" and "+the_list[-1]
    else:
        return ", ".join(the_list[:2])+", ... and "+the_list[-1]

def get_valid_filename(s):
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)

def get_digits(x, digits=6):
    prtd = max(digits, math.ceil(math.log10(abs(x))))
    prtd -= math.ceil(math.log10(abs(x)))
    prtd = min(digits, prtd)
    return prtd

def format_digits(x, digits=6):
    if isinstance(x, list):
        return format_digits(max([get_digits(i, digits=digits) for i in x if i is not None]), digits=digits)

    elif x is None:
        return 0
    elif x == 0 or not math.isfinite(x):
        return x
    else:
        prtd = max(digits, math.ceil(math.log10(abs(x))))
        prtd -= math.ceil(math.log10(abs(x)))
        prtd = min(digits, prtd)
        return f'1.{prtd}f'

def get_h0_equal_means(columns):
    if len(columns) == 1:
        return ''
    elif len(columns) > 4:
        return f'$H_0: \\bar{{X}}_{{{columns[0]}}} = \\bar{{X}}_{{{columns[1]}}} = \\bar{{X}}_{{...}} = \\bar{{X}}_{{{columns[-1]}}}$'
    else:
        result = f'$H_0: \\bar{{X}}_{{{columns[0]}}}$'
        for col in columns[1:]:
            result += f'$ = \\bar{{X}}_{{{col}}}$'
        return result
    
def get_distribution(dist=None):
    distributions = {
        "norm": "Normal (Gaussian)",
        "alpha": "Alpha",
        "anglit": "Anglit",
        "arcsine": "Arcsine",
        "beta": "Beta",
        "betaprime": "Beta Prime",
        "bradford": "Bradford",
        "burr": "Burr",
        "cauchy": "Cauchy",
        "chi": "Chi",
        "chi2": "Chi-squared",
        "cosine": "Cosine",
        "dgamma": "Double Gamma",
        "dweibull": "Double Weibull",
        "erlang": "Erlang",
        "expon": "Exponential",
        "exponweib": "Exponentiated Weibull",
        "exponpow": "Exponential Power",
        "fatiguelife": "Fatigue Life (Birnbaum-Sanders)",
        "foldcauchy": "Folded Cauchy",
        "f": "F (Snecdor F)",
        "fisk": "Fisk",
        "foldnorm": "Folded Normal",
        "frechet_r": "Frechet Right Sided, Extreme Value Type II",
        "frechet_l": "Frechet Left Sided, Weibull_max",
        "gamma": "Gamma",
        "gausshyper": "Gauss Hypergeometric",
        "genexpon": "Generalized Exponential",
        "genextreme": "Generalized Extreme Value",
        "gengamma": "Generalized gamma",
        "genlogistic": "Generalized Logistic",
        "genpareto": "Generalized Pareto",
        "genhalflogist": "Generalized Half Logistic",
        "gilbrat": "Gilbrat",
        "gompertz": "Gompertz (Truncated Gumbel)",
        "gumbel_l": "Left Sided Gumbel, etc.",
        "gumbel_r": "Right Sided Gumbel",
        "halfcauchy": "Half Cauchy",
        "halflogistic": "Half Logistic",
        "halfnorm": "Half Normal",
        "hypsecant": "Hyperbolic Secant",
        "invgamma": "Inverse Gamma",
        "invnorm": "Inverse Normal",
        "invweibull": "Inverse Weibull",
        "johnsonsb": "Johnson SB",
        "johnsonsu": "Johnson SU",
        "laplace": "Laplace",
        "logistic": "Logistic",
        "loggamma": "Log-Gamma",
        "loglaplace": "Log-Laplace (Log Double Exponential",
        "lognorm": "Log-Normal",
        "lomax": "Lomax (Pareto of the second kind)",
        "maxwell": "Maxwell",
        "mielke": "Mielke's Beta-Kappa",
        "nakagami": "Nakagami",
        "ncx2": "Non-central chi-squared",
        "ncf": "Non-central F",
        "nct": "Non-central Student's T",
        "pareto": "Pareto",
        "powerlaw": "Power-function",
        "powerlognorm": "Power log normal",
        "powernorm": "Power normal",
        "rdist": "R distribution",
        "reciprocal": "Reciprocal",
        "rayleigh": "Rayleigh",
        "rice": "Rice",
        "recipinvgauss": "Reciprocal Inverse Gaussian",
        "semicircular": "Semicircular",
        "t": "Student's T",
        "triang": "Triangular",
        "truncexpon": "Truncated Exponential",
        "truncnorm": "Truncated Normal",
        "tukeylambda": "Tukey-Lambda",
        "uniform": "Uniform",
        "vonmises": "Von-Mises (Circular)",
        "wald": "Wald",
        "weibull_min": "Minimum Weibull (see Frechet)",
        "weibull_max": "Maximum Weibull (see Frechet)",
        "wrapcauchy": "Wrapped Cauchy",
        "ksone": "Kolmogorov-Smirnov one-sided (no st",
        "kstwobign": "Kolmogorov-Smirnov two-sided test for Large N",
    }

    return distributions.get(dist, None)


import pandas as pd
import scipy.stats as stats

import matplotlib.pyplot as plt

from bluebelt.core import graphs

import bluebelt.statistics.hypothesis_testing

import bluebelt.helpers.mpl_format as mpl_format

import bluebelt.styles

class WeekDay():
    
    """
    Show the data distribution per weekday
    series: pandas.Series
    """

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=[self.series.index.weekday, self.series.index.day_name()], data=self.series.values, name=self.series.name).sort_index(level=0).droplevel(0)
        self.data = [self.series[day].dropna().values for day in self.series.index.unique()]
        self.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(pd.DataFrame(self.data).T).passed
        self.equal_variance = bluebelt.statistics.hypothesis_testing.EqualVariance(pd.DataFrame(self.data).T).passed
        
        
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variance={self.equal_variance})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} week day distribution', **kwargs)
        
class MonthDay():
    
    """
    Show the data distribution per day of the month
    series: pandas.Series
    """

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.day, data=self.series.values, name=self.series.name).sort_index()
        self.data = [self.series[day].dropna().values for day in self.series.index.unique()]
        self.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(pd.DataFrame(self.data).T).passed
        self.equal_variance = bluebelt.statistics.hypothesis_testing.EqualVariance(pd.DataFrame(self.data).T).passed
                
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variance={self.equal_variance})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} month day distribution', **kwargs) 

class Week():
    
    """
    Show the data distribution per weekday
    series: pandas.Series
    """

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.isocalendar().week, data=self.series.values, name=self.series.name).sort_index()
        self.data = [self.series[week].dropna().values for week in self.series.index.unique()]
        self.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(pd.DataFrame(self.data).T).passed
        self.equal_variance = bluebelt.statistics.hypothesis_testing.EqualVariance(pd.DataFrame(self.data).T).passed

    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variance={self.equal_variance})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} week distribution', **kwargs)
        
class Month():
    
    """
    Show the data distribution per month
    series: pandas.Series
    """

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.month, data=self.series.values, name=self.series.name).sort_index()
        self.data = [self.series[month].dropna().values for month in self.series.index.unique()]
        self.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(pd.DataFrame(self.data).T).passed
        self.equal_variance = bluebelt.statistics.hypothesis_testing.EqualVariance(pd.DataFrame(self.data).T).passed
                
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variance={self.equal_variance})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} month distribution', **kwargs)
        
class Year():
    
    """
    Show the data distribution per year
    series: pandas.Series
    """

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.year, data=self.series.values, name=self.series.name).sort_index()
        self.data = [self.series[year].dropna().values for year in self.series.index.unique()]
        self.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(pd.DataFrame(self.data).T).passed
        self.equal_variance = bluebelt.statistics.hypothesis_testing.EqualVariance(pd.DataFrame(self.data).T).passed
                
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variance={self.equal_variance})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} year distribution')

def _datetime_plot(plot_obj, **kwargs):
        
    title = kwargs.pop('title', f'{plot_obj.name} datetime distribution')
    style = kwargs.pop('style', bluebelt.styles.paper)
    
    path = kwargs.pop('path', None)        
    
    # prepare figure
    fig, ax = plt.subplots(**kwargs)

    boxplot = ax.boxplot(plot_obj.data)

    for n, box in enumerate(boxplot['boxes']):
        # add style if any is given
        box.set(**style.graphs.boxplot.boxplot.get('boxes', {}))
        
    # title
    ax.set_title(title, **style.graphs.boxplot.title)

    # labels        
    ax.set_xticklabels(plot_obj.series.index.unique())

    plt.gcf().subplots_adjust(right=0.8)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


import pandas as pd

import bluebelt.styles
import bluebelt.helpers.mpl_format as mpl_format

import warnings

def area(series, ax, **kwargs):

    # handle kwargs
    style = kwargs.pop('style', bluebelt.styles.paper)
    title = kwargs.pop('title', f'{series.name} area plot')

    # area plot
    ax.stackplot(series.index, series.values, **style.graphs.area.stackplot)
    ax.plot(series, **style.graphs.area.plot)
    
    # labels
    ax.set_title(title, **style.graphs.area.title)
    
    # set x axis locator
    mpl_format.axisformat(ax, series)
    ax.set_ylim(0,)

def hist(series, ax, **kwargs):

    # handle kwargs
    xstyle = kwargs.pop('style', bluebelt.styles.paper)
    title = kwargs.pop('title', f'{series.name} histogram')
    fit_normal_distribution = kwargs.pop('fit_normal_distribution', False)

    # histogram
    ax.hist(series, **style.graphs.hist.hist)
    ax.set_yticks([])

    # fit a normal distribution to the data
    if fit_normal_distribution:
        norm_mu, norm_std = stats.norm.fit(series.dropna())
        xlims = ax.get_xlim()
        pdf_x = np.linspace(xlims[0], xlims[1], 100)
        ax.plot(pdf_x, stats.norm.pdf(pdf_x, norm_mu, norm_std), **style.graphs.hist.plot)

    # labels
    ax.set_title(title, **style.graphs.area.title)

def boxplot(series, ax, **kwargs):

    # get data
    if isinstance(series, (pd.Series, pd.DataFrame))and series.isnull().values.any():
        warnings.warn('the series contains Null values which will be removed', Warning)
        series = series.dropna()

    style = kwargs.pop('style', bluebelt.styles.paper)
    title = kwargs.pop('title', None)


        
    boxplot = ax.boxplot(series)

    for n, box in enumerate(boxplot['boxes']):
        # add style if any is given
        box.set(**style.graphs.boxplot.boxplot.get('boxes', {}))
        
    # title
    ax.set_title(title, **style.graphs.boxplot.title)


def meeting_effectiveness():
    result = """
    ### complience
    1. The meeting starts on time.
    2. All participants are present or have a good reason for absence (illness, vacation, etc ...).
    3. All participants are on time.
    4. All participants are well prepared.
    5. All relevant information is distributed before the meeting.
    6. Terms of Reference are used to include the participants in the goal of the meeting.
    7. An actions and decisions list is used.
    8. The meeting ends on time.

    ### connection
    1. The chairman is aware of all relevant issues and knows which ones need to be discussed.
    2. There is a good balance between main and side issues.
    3. All items on the agenda are sufficiently covered.
    4. All completed actions are discussed and new actions are assigned.
    5. The chairman pays attention to actions that have not been completed.
    6. All participants have a proactive attitude (solutions instead of problems).
    7. Strikingly good or bad results are discussed in detail.
    8. Discussions focus on the performance issues that have the greatest impact.

    ### completion
    1. The planning and realization are discussed in detail.
    2. The root causes of any gap between planning and realization are determined.
    3. Responsibles (R in RASCI) can explain the differences between planning and realization.
    4. Discussions focus on the performance issues that have the greatest impact.
    5. KPIs are updated when the needs of the business have changed.
    6. The chairman carefully weighs new targets in order to continue to improve.
    7. Undesirable behavior is corrected and managers coach and discuss as part of performance management.
    8. Lessons learned are communicated, best practices are shared between departments and teams.
    """
    return result

def mef():
    mef = {
        'complience': [
            'The meeting starts on time.',
            'All participants are present or have a good reason for absence (illness, vacation, etc ...).',
            'All participants are on time.',
            'All participants are well prepared.',
            'All relevant information is distributed before the meeting.',
            'Terms of Reference are used to include the participants in the goal of the meeting.',
            'An actions and decisions list is used.',
            'The meeting ends on time.',
            ],
        'connection': [
            'The chairman is aware of all relevant issues and knows which ones need to be discussed.',
            'There is a good balance between main and side issues.',
            'All items on the agenda are sufficiently covered.',
            'All completed actions are discussed and new actions are assigned.',
            'The chairman pays attention to actions that have not been completed.',
            'All participants have a proactive attitude (solutions instead of problems).',
            'Strikingly good or bad results are discussed in detail.',
            'Discussions focus on the performance issues that have the greatest impact.',
            ],
        'completion': [
            'The planning and realization are discussed in detail.',
            'The root causes of any gap between planning and realization are determined.',
            'Responsibles (R in RASCI) can explain the differences between planning and realization.',
            'Discussions focus on the performance issues that have the greatest impact.',
            'KPIs are updated when the needs of the business have changed.',
            'The chairman carefully weighs new targets in order to continue to improve.',
            'Undesirable behavior is corrected and managers coach and discuss as part of performance management.',
            'Lessons learned are communicated, best practices are shared between departments and teams.',
        ]
    }


import bluebelt.helpers.defaults as defaults

# boxplot | pyplot.boxplot
boxplot = {
    'boxes': {
        'hatch': '//',
        },
}

# mean axvline | pyplot.axvline
mean_axvline = {
    'color': defaults.blue,
    'linestyle': 'dotted',
    'linewidth': 1,
}

# population mean axvline | pyplot.axvline
popmean_axvline = {
    'color': defaults.red,
    'linestyle': 'dashed',
    'linewidth': 1,
    'zorder': 99,
}

# ci mean horizontal line | pyplot.plot
ci_mean_plot = {
    'color': defaults.black,
    'zorder': 90,
}

# ci mean vertical lines (whiskers) | pyplot.plot
ci_mean_axvline = {
    'color': defaults.black,
    'zorder': 90,
}

# ci mean scatter (mean point) : pyplot.scatter
ci_mean_scatter = {
    'color': defaults.black,
    'zorder': 90,
}

# result text | pyplot.text
result_text = {
    'backgroundcolor': defaults.white,
    'va': 'top',
    'ha': 'left', 
}

# text | pyplot.text
text = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'center',
    'zorder': 10,
}

# text | pyplot.text
text_ci_min = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'right',
    'zorder': 0,
}

# text | pyplot.text
text_ci_max = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'left',
    'zorder': 0, 
}


ci_fill_between = {
    'alpha': 1,
    #'color': None,
    'edgecolor': defaults.blue,
    'facecolor': None, # color name, hex or None
    'hatch': '\\\\\\\\',
    'linewidth': 0,
}

# standard deviation axvline | pyplot.plot
ci_axvline = {
    'color': defaults.blue,
    'linestyle': 'dotted',
    'linewidth': 1,
    'zorder': 10,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
    'size': defaults.small
}

# plot title | pyplot.set_title
suptitle = {
    'ha': 'left',
    'size': defaults.medium
}



import pandas as pd
import numpy as np
import math

import matplotlib.pyplot as plt

from bluebelt.core.checks import check_kwargs

import bluebelt.core.decorators


def resolution_methods(cls):

    def sum(self):
        self.result = self.result.sum()
        self.func = 'sum'
        return self

    def mean(self):
        self.result = self.result.mean()
        self.func = 'mean'
        return self

    def var(self):
        self.result = self.result.var()
        self.func = 'variance'
        return self

    def std(self):
        self.result = self.result.std()
        self.func = 'standard deviation'
        return self

    def min(self):
        self.result = self.result.min()
        self.func = 'min'
        return self

    def max(self):
        self.result = self.result.max()
        self.func = 'max'
        return self
    
    def count(self):
        self.result = self.result.count()
        self.func = 'count'
        return self
    
    def value_range(self):
        self.result = self.result.apply(lambda x: x.max() - x.min())
        self.func = 'value range'
        return self

    def subsize_count(self, count=3, size=1):
        """
        Count the number of times a list of <count> items with size <size> fit in the groupby object (which is a pandas Series)
        e.g.
        groupby object: pd.Series([10, 8, 3, 3, 5])
        count = 3
        size = 1

        returns 9

        step 0: (3, 3, 5, 8, 10)
        step 1: (3, 3, 4, 7, 9)
        step 2: (3, 3, 3, 6, 8)
        step 3: (2, 3, 3, 5, 7)
        step 4: (2, 2, 3, 4, 6)
        step 5: (2, 2, 2, 3, 5)
        step 6: (1, 2, 2, 2, 4)
        step 7: (1, 1, 1, 2, 3)
        step 8: (0, 1, 1, 1, 2)
        step 9: (0, 0, 0, 1, 1)

        """
        self.result = self.result.apply(lambda x: _subsize_count(series=x, count=count, size=size))
        self.func = f'count subsize (count={count}, size={size})'
        return self

    def subseries_count(self, subseries=None, **kwargs):
        """
        Count the number of times a subseries of wil fit in the groupby object (which is a pandas Series)
        e.g.
        groupby object: pd.Series([10, 8, 3, 3, 5])
        subseries: pd.Series([1, 2, 3])

        returns 4

        step 0: (3, 3, 5, 8, 10)
        step 1: (3, 3, 4, 6, 7)
        step 2: (3, 3, 3, 4, 4)
        step 3: (2, 3, 3, 2, 1)
        step 4: (1, 1, 0, 2, 1)

        """
        self.result = self.result.apply(lambda x: _subseries_count(series=x, subseries=subseries, **kwargs))
        self.func = f'count subseries (subseries={list(subseries)})'
        return self

    def plot(self, **kwargs):

        if self.func == '':
            series = self.result.sum()
            func = 'sum'
        else:
            series = self.result
            func = self.func

        style = kwargs.pop('style', bluebelt.styles.paper)
        path = kwargs.pop('path', None)

        fig, ax = plt.subplots(**kwargs)

        # observations
        ax.plot(series, **style.resolution.plot)

        # labels
        ax.set_title(f'{self.series.name} grouped', **style.resolution.title)
        ax.set_xlabel(self.index)
        ax.set_ylabel(func)

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

    setattr(cls, 'sum', sum)
    setattr(cls, 'mean', mean)
    setattr(cls, 'var', var)
    setattr(cls, 'std', std)
    setattr(cls, 'min', min)
    setattr(cls, 'max', max)
    setattr(cls, 'count', count)
    setattr(cls, 'value_range', value_range)
    setattr(cls, 'subsize_count', subsize_count)
    setattr(cls, 'subseries_count', subseries_count)
    setattr(cls, 'plot', plot)
    
    return cls

@bluebelt.core.decorators.class_methods
@resolution_methods
class Resample():
    """
    Resample as series and apply a specific function.
        arguments
        series: pandas.Series
        rule: str
            the resampling rule
        any pandas.Series.resample argunment

        Apply one of the following functions:
            .sum()
            .mean()
            .min()
            .max()
            .std()
            .value_range()
            .count()
            .subsize_count()
            .subseries_count()

        e.g. series.blue.data.resample(rule="1W").sum()
    """
    def __init__(self, series, rule="1W", **kwargs):

        check_kwargs(locals())

        # fetch resample kwargs
        resample_kwargs = {
            'rule': rule,
            'axis': kwargs.pop('axis', 0),
            'closed': kwargs.pop('closed', None),
            'label': kwargs.pop('label', None),
            'convention': kwargs.pop('convention', 'start'),
            'kind': kwargs.pop('kind' ,None),
            'loffset': kwargs.pop('loffset', None),
            'base': kwargs.pop('base', None),
            'on': kwargs.pop('on', None),
            'level': kwargs.pop('level', None),
            'origin': kwargs.pop('origin', 'start_day'),
            'offset': kwargs.pop('offset', None),
        }

        self.series = series
        self.index = rule
        self.func = ''
        self.nrows = self.series.size
        self.resample_kwargs = resample_kwargs
        self.kwargs = kwargs
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.result = self.series.resample(**self.resample_kwargs)

    def __str__(self):
        return ""
    
    def __repr__(self):
        return self.result.__repr__()

@bluebelt.core.decorators.class_methods
@resolution_methods
class GroupByDatetimeIndex():
    """
    Group a series by DateTime index and apply a specific function.
        arguments
        series: pandas.Series
        index: str
            a string with date-time keywords that can be parsed to group the index
            keywords:
                weekday_name -> %a  # abbreviated weekday name
                week_day -> %u      # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                weekday -> %u       # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                iso_week -> %V      # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                isoweek -> %V       # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                monthname -> %b     # abbreviated month name
                month_name -> %b    # abbreviated month name
                monthday -> %d      # day of the month (01 to 31)
                month_day -> %d     # day of the month (01 to 31)
                isoyear -> %G       # 4-digit year corresponding to the ISO week number (see %V).
                iso_year -> %G      # 4-digit year corresponding to the ISO week number (see %V).
                yearday -> %j       # day of the year (001 to 366)
                year_day -> %j      # day of the year (001 to 366)
                year -> %Y          # year including the century
                month -> %m         # month (01 to 12)
                week -> %W          # week number of the current year, starting with the first Monday as the first day of the first week
                day -> %d           # day of the month (01 to 31)
                hour -> %H          # hour, using a 24-hour clock (00 to 23)
                minute -> %M        # minute
                second -> %S        # second
                time -> %T          # current time, equal to %H:%M:%S
            
            default value "iso_year-iso_week"

            e.g. "iso_year-Wiso_week" will parse to "%G-W%V" which will print "2021-W01" for the first week of 2021

        Apply one of the following functions:
            .sum()
            .mean()
            .min()
            .max()
            .std()
            .value_range()
            .count()
            .subsize_count()
            .subseries_count()

        e.g. series.blue.data.group_index(rule="1W").sum()
        
    """
    def __init__(self, series, index="iso_year-iso_week", **kwargs):

        self.series = series
        self.index = index
        self.nrows = self.series.size
        self.func = ''
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.result = self.series.groupby(self.series.index.strftime(_strf(self.index)))

    def __str__(self):
        return ""
    
    def __repr__(self):
        return self.result.__repr__()

def _subsize_count(series, count=3, size=1):
    series = pd.Series(series)/size
    result = series.sum()*count
    for i in range(count, 0, -1):
        result = min(result, math.floor(series.nsmallest(len(series) - count + i).sum() / i))
    return result

def _subseries_count(series, subseries=None, **kwargs):
    series = pd.Series(series)
    subseries = pd.Series(subseries)
    result=series.sum()*subseries.sum()
    for i in range(len(subseries), 0, -1):
        result = min(result, math.floor(series.nsmallest(len(series) - len(subseries) + i).sum() / subseries.nsmallest(i).sum()))
    return result

def _strf(string):
    strf_dict = {'weekday_name': '%a', # abbreviated weekday name
                'week_day': '%u', # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                'weekday': '%u', # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                'iso_week': '%V', # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                'isoweek': '%V', # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                'monthname': '%b', # abbreviated month name
                'month_name': '%b', # abbreviated month name
                'monthday': '%d', # day of the month (01 to 31)
                'month_day': '%d', # day of the month (01 to 31)
                'isoyear': '%G', # 4-digit year corresponding to the ISO week number (see %V).
                'iso_year': '%G', # 4-digit year corresponding to the ISO week number (see %V).
                'yearday': '%j', # day of the year (001 to 366)
                'year_day': '%j', # day of the year (001 to 366)
                'year': '%Y', # year including the century
                'month': '%m', # month (01 to 12)
                'week': '%W', # week number of the current year, starting with the first Monday as the first day of the first week
                'day': '%d', # day of the month (01 to 31)
                'hour': '%H', # hour, using a 24-hour clock (00 to 23)
                'minute': '%M', # minute
                'second': '%S', # second
                'time': '%T', # current time, equal to %H:%M:%S
                }
    for key, value in strf_dict.items():
        string = string.replace(key, value)
    
    return string




# process performance analysis
import copy

import pandas as pd
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

import plotly.graph_objects as go

import bluebelt.statistics.std

import bluebelt.analysis.ci as ci
from bluebelt.analysis.subgroups import get_subgroups
import bluebelt.statistics.hypothesis_testing as hypothesis_testing

import bluebelt.helpers.defaults as defaults
import bluebelt.helpers.histogram
import bluebelt.helpers.boxplot
import bluebelt.helpers.mpl_format as mpl_format
import bluebelt.helpers.text

import bluebelt.core.decorators
import bluebelt.styles

# process performance analysis
@bluebelt.core.decorators.class_methods
class Summary():
        
    def __init__(self, series, whis=1.5, **kwargs):

        # check arguments
        if not isinstance(series, pd.Series):
            raise ValueError('series is not a Pandas Series')
        
        self.series = series.dropna()
        self.whis = whis
        self.calculate(whis)
        
    def __str__(self):
        str_mean="mean:"
        str_ci_mean="CI mean:"
        str_std="standard deviation:"
        str_min="minimum"
        str_q1="1st quantile:"
        str_median="median:"
        str_q3="3rd quantile:"
        str_max="maximum"
        str_ci_median="CI median:"
        str_ad_test="Anderson-Darling test"
        
        fill = 32
        return (f'{str_mean:{fill}}{self.mean:1.2f}\n' +
                f'{str_ci_mean:{fill}}{self.ci_mean[0]:1.2f}-{self.ci_mean[1]:1.2f}\n' +
                f'{str_std:{fill}}{self.std:1.2f}\n' +
                f'{str_min:{fill}}{self.min:1.2f}\n' +
                f'{str_q1:{fill}}{self.q1:1.2f}\n' +
                f'{str_median:{fill}}{self.median}\n' +
                f'{str_q3:{fill}}{self.q3:1.2f}\n' +
                f'{str_max:{fill}}{self.max:1.2f}\n' +
                f'{str_ci_median:{fill}}{self.ci_median[0]:1.2f}-{self.ci_median[1]:1.2f}\n' +
                f'{str_ad_test:{fill}}A={self.ad_test.statistic:1.2f}, p-value={self.ad_test.p_value:1.2f}')
                

    def __repr__(self):
        return (f'{self.__class__.__name__}(mean={self.mean:1.1f}, std={self.std:1.1f}, min={self.min:1.1f}, q1={self.q1:1.1f}, median={self.median:1.1f}, q3={self.q3:1.1f}, max={self.max:1.1f})')
    
    def calculate(self, whis=1.5):
        
        self.mean = self.series.mean()
        self.ci_mean = ci.ci_mean(self.series)
        self.std = self.series.std()
        self.min = self.series.min()
        self.q1 = self.series.quantile(q=0.25)
        self.median = self.series.median()
        self.q3 = self.series.quantile(q=0.75)
        self.max = self.series.max()
        self.ci_median = ci.ci_median(self.series)
        self.ad_test = hypothesis_testing.AndersonDarling(self.series)
        self.boxplot_quantiles = self.get_boxplot_quantiles(whis=whis)
        self.boxplot_outliers = self.get_boxplot_outliers(whis=whis)
        
        
    def get_boxplot_quantiles(self, whis=1.5):
        # get the matplotlib boxplot quantiles
        # sort of direct copy from the matplotlib library

        q1, median, q3 = np.percentile(self.series, [25, 50, 75])

        if np.isscalar(whis):
            iqr = q3 - q1
            loval = q1 - (whis * iqr)
            hival = q3 + (whis * iqr)
        else:
            loval, hival = np.percentile(self.series, sorted(whis))

        # get high extreme
        wiskhi = self.series[self.series <= hival]
        if len(wiskhi) == 0 or np.max(wiskhi) < q3:
            high = q3
        else:
            high = np.max(wiskhi)

        # get low extreme
        wisklo = self.series[self.series >= loval]
        if len(wisklo) == 0 or np.min(wisklo) > q1:
            low = q1
        else:
            low = np.min(wisklo)
        
        return (low, q1, q3, high)

    def get_boxplot_outliers(self, whis=1.5):
        low, q1, q3, high = self.get_boxplot_quantiles(whis=whis)

        return self.series[self.series < low].append(self.series[self.series > high]).values
    
    def plot(self, **kwargs):

        style = kwargs.get('style', bluebelt.styles.paper)
        kwargs.pop('style', None)
        path = kwargs.pop('path', None)

        # prepare figure
        fig = plt.figure(constrained_layout=False, **kwargs)
        gridspec = fig.add_gridspec(nrows=4, ncols=1, height_ratios=[8,3,3,3], wspace=0, hspace=0)
        ax1 = fig.add_subplot(gridspec[0, 0], zorder=50)
        ax2 = fig.add_subplot(gridspec[1, 0], zorder=40)
        ax3 = fig.add_subplot(gridspec[2, 0], zorder=30)
        ax4 = fig.add_subplot(gridspec[3, 0], zorder=20)

        # 1. histogram ############################################
        ax1.hist(self.series, **style.summary.histogram)

        # get current limits
        xlims = ax1.get_xlim()
        ylims = ax1.get_ylim()

        # fit a normal distribution to the data
        norm_mu, norm_std = stats.norm.fit(self.series)
        pdf_x = np.linspace(xlims[0], xlims[1], 100)
        ax1.plot(pdf_x, stats.norm.pdf(pdf_x, norm_mu, norm_std), **style.summary.normal_plot)

        # plot standard deviation
        if (self.mean-self.std) > self.min:
            std_area = np.linspace(self.mean-self.std, self.mean, 100)
            std_line_x = self.mean-self.std
            std_text_x = self.mean-self.std*0.5
        else:
            std_area = np.linspace(self.mean, self.mean+self.std, 100)
            std_line_x = self.mean+self.std
            std_text_x = self.mean+self.std*0.5

        ax1.fill_between(std_area, stats.norm.pdf(std_area, norm_mu, norm_std), 0, **style.summary.std_fill_between)
        
        ax1.axvline(x=self.mean, ymin=0, ymax=1, **style.summary.std_axvline)
        ax1.axvline(x=std_line_x, ymin=0, ymax=1, **style.summary.std_axvline)
        
        ax1.plot((std_line_x, self.mean), (ylims[1]*0.9, ylims[1]*0.9), **style.summary.std_axhline)
        ax1.text(std_text_x, ylims[1]*0.9, r'$\sigma = $'+f'{self.std:1.2f}', **style.summary.std_text)
        
        # plot AD test results
        if self.mean > (self.max + self.min) / 2:
            ad_x = 0.02
            ad_align = 'left'
        else:
            ad_x =  0.98
            ad_align = 'right'

        ad_text = r'$P_{AD, normal}: $'+f'{self.ad_test.p_value:1.4f}'
        
        ax1.text(ad_x, 0.9, ad_text, transform=ax1.transAxes, va='center', ha=ad_align)

        # reset limits
        ax1.set_xlim(xlims)
        ax1.set_ylim(ylims)

        # 2. box plot ############################################
        boxplot = ax2.boxplot(self.series, vert=False, widths=0.3, whis=self.whis)
        
        for box in boxplot['boxes']:
            # add style if any is given
            box.set(**style.summary.boxplot.get('boxes', {}))

        ax2.set_xlim(xlims)
        ax2.set_ylim(0.7,1.7)
        
        ax2_xticks = bluebelt.helpers.boxplot.get_boxplot_quantiles(self.series, whis=1.5)
        ax2.set_xticks(ax2_xticks)
        ax2.xaxis.set_major_formatter(FormatStrFormatter(f'%{bluebelt.helpers.text.format_digits(ax2_xticks, digits=3)}'))
        

        

        #for tick in ax2.get_xticklabels():
        #    tick.set_horizontalalignment('left')

        #######################################################
        # CI for the median
        #######################################################

        ax3.plot([self.ci_median[0], self.ci_median[1]], [1,1], **style.summary.ci_mean_plot)
        ax3.axvline(x=self.ci_median[0], ymin=0.15, ymax=0.45, **style.summary.ci_mean_axvline)
        ax3.axvline(x=self.ci_median[1], ymin=0.15, ymax=0.45, **style.summary.ci_mean_axvline)
        ax3.scatter([self.median],[1], **style.summary.ci_mean_scatter)
        ax3.set_xlim(xlims)
        ax3_xticks = [self.ci_median[0], self.ci_median[1]]
        
        
        ax3.set_ylim(0.7,1.7)
        
        ci_median_x = 0.02 if self.median>(self.max+self.min)/2 else 0.98
        ci_median_align = 'left' if self.median>(self.max+self.min)/2 else 'right'
        
        ax3.text(ci_median_x, 0.1, r'$CI_{median}$', transform=ax3.transAxes, va='bottom', ha=ci_median_align)

        ax3.set_xticks(ax3_xticks)
        ax3.xaxis.set_major_formatter(FormatStrFormatter(f'%{bluebelt.helpers.text.format_digits(ax3_xticks, digits=3)}'))


        #######################################################
        # CI for the mean
        #######################################################

        ax4.plot([self.ci_mean[0], self.ci_mean[1]], [1,1], **style.summary.ci_median_plot)
        ax4.axvline(x=self.ci_mean[0], ymin=0.15, ymax=0.45, **style.summary.ci_median_axvline)
        ax4.axvline(x=self.ci_mean[1], ymin=0.15, ymax=0.45, **style.summary.ci_median_axvline)
        ax4.scatter([self.mean],[1], **style.summary.ci_median_scatter)
        ax4.set_xlim(xlims)
        ax4_xticks = [self.ci_mean[0], self.ci_mean[1]]
        ax4.set_ylim(0.7,1.7)

        ci_mean_x = 0.02 if self.mean>(self.max+self.min)/2 else 0.98
        ci_mean_align = 'left' if self.mean>(self.max+self.min)/2 else 'right'
        
        ax4.text(ci_mean_x, 0.1, r'$CI_{mean}$', transform=ax4.transAxes, va='bottom', ha=ci_mean_align)

        # drop lines
        ax2.axvline(self.mean, ymin=0, ymax=2, **style.summary.drop_axvline)
        ax3.axvline(self.mean, ymin=0, ymax=1.7, **style.summary.drop_axvline)
        ax4.axvline(self.mean, ymin=0.3, ymax=1.7, **style.summary.drop_axvline)
        
        ax2.axvline(self.median, ymin=0, ymax=0.3, **style.summary.drop_axvline)
        ax3.axvline(self.median, ymin=0.3, ymax=1.7, **style.summary.drop_axvline)
        


        ax4.set_xticks(ax4_xticks)
        ax4.xaxis.set_major_formatter(FormatStrFormatter(f'%{bluebelt.helpers.text.format_digits(ax4_xticks, digits=3)}'))

        ax1.set_title(f'graphical summary for {self.series.name}', **style.summary.title)

        ax1.set_yticks([])
        ax2.set_yticks([])
        ax3.set_yticks([])
        ax4.set_yticks([])

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

@bluebelt.core.decorators.class_methods
class ControlChart():
    
    def __init__(self, series, **kwargs):

        # check arguments
        if not isinstance(series, pd.Series):
            raise ValueError('series is not a Pandas Series')
                
        self.series = series
        self.calculate()
        
    def __str__(self):
        str_mean="mean:"
        str_std="standard deviation:"
        str_ucl="upper control limit:"
        str_lcl=f"lower control limit:"
        str_outlier_count=f"outliers:"

        fill = 32
        return (f'{str_mean:{fill}}{self.mean:1.2f}\n' +
                f'{str_std:{fill}}{self.std:1.2f}\n' +
                f'{str_ucl:{fill}}{self.ucl:1.2f}\n' +
                f'{str_lcl:{fill}}{self.lcl:1.2f}\n' +
                f'{str_outlier_count:{fill}}{self.outlier_count}\n') 

    def __repr__(self):
        return (f'{self.__class__.__name__}(mean={self.mean:1.1f}, std={self.std:1.1f}, UCL={self.ucl:1.1f}, LCL={self.lcl:1.1f}, outlier_count={self.outlier_count:1.0f})')
    
    def calculate(self):
        mean = self.series.mean()
        std = self.series.std()
        ucl = mean + std * 3
        lcl = mean - std * 3
        outliers = self.series[(self.series > ucl) | (self.series < lcl)]
        
        self.mean = mean
        self.std = std
        self.ucl = ucl
        self.lcl = lcl
        self.outliers = outliers
        self.outlier_count = outliers.shape[0]
        
    def plot(self, **kwargs):
        
        style = kwargs.get('style', bluebelt.styles.paper)
        kwargs.pop('style', None)

        path = kwargs.pop('path', None)
        
        fig, ax = plt.subplots(**kwargs)

        # observations
        ax.plot(self.series, **style.control_chart.plot)

        # observations white trail
        ax.plot(self.series, **style.control_chart.plot_background)

        # control limits
        ax.axhline(self.ucl, **style.control_chart.ucl_axhline)
        ax.axhline(self.lcl, **style.control_chart.lcl_axhline)

        ylim = ax.get_ylim() # get limits to set them back later
        xlim = ax.get_xlim()
        
        ax.fill_between(xlim, self.ucl, ylim[1], **style.control_chart.ucl_fill_between)
        ax.fill_between(xlim, self.lcl, ylim[0], **style.control_chart.lcl_fill_between)

        # outliers
        ax.plot(self.outliers, **style.control_chart.outlier_background)
        ax.plot(self.outliers, **style.control_chart.outlier)
        
        # mean
        ax.axhline(self.mean, **style.control_chart.mean_axhline)

        # text boxes for mean, UCL and LCL
        ax.text(xlim[1], self.mean, f' mean = {self.mean:1.2f}', **style.control_chart.mean_text)
        ax.text(xlim[1], self.ucl, f' UCL = {self.ucl:1.2f}', **style.control_chart.ucl_text)
        ax.text(xlim[1], self.lcl, f' LCL = {self.lcl:1.2f}', **style.control_chart.lcl_text)

        # reset limits
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)

        # labels
        ax.set_title(f'control chart of {self.series.name}', **style.control_chart.title)
        #ax.set_xlabel(self.series.index.name)
        ax.set_ylabel('value')

        # set x axis locator
        mpl_format.axisformat(ax, self.series)
        
        plt.gcf().subplots_adjust(right=0.8)

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig
    
    def plotly(self, **kwargs):
        
        layout = go.Layout(
            title=f'control chart of {self.series.name}',    
            plot_bgcolor=f'rgba{defaults.white+(1,)}',
            hovermode="closest",
            xaxis=dict(
                title=self.series.index.name,
                linecolor=f'rgba{defaults.black+(1,)}',
                zeroline=False,
                ticks="outside",
                tickwidth=1,
                tickcolor=f'rgba{defaults.black+(1,)}',
                ticklen=5,
                showgrid=True,
                gridwidth=0.5,
                gridcolor=f'rgba{defaults.light_grey+(1,)}',
                mirror=True,
            ),
            yaxis=dict(
                title="value",
                linecolor=f'rgba{defaults.black+(1,)}',
                zeroline=False,
                ticks="outside",
                tickwidth=1,
                tickcolor=f'rgba{defaults.black+(1,)}',
                ticklen=5,
                showgrid=True,
                gridwidth=0.5,
                gridcolor=f'rgba{defaults.light_grey+(1,)}',
                mirror=True,
            ),
        )
        
        data = [go.Scatter(
                        x=self.outliers.index,
                        y=self.outliers.values,
                        marker=dict(
                            color=f'rgba{defaults.white+(1,)}',
                            line=dict(width=1, color=f'rgba{defaults.black+(1,)}'),
                            size=17,
                        ),
                        mode='markers',
                        showlegend=False,
                        name='outliers'
                    ),
                go.Scatter(
                        x=self.series.index,
                        y=self.series.values,
                        line=dict(
                                width=1,
                                color=f'rgba{defaults.blue+(1,)}',
                            ),
                        marker=dict(
                            color=f'rgba{defaults.black+(1,)}',
                            size=9,
                        ),
                        mode='lines+markers',
                        showlegend=False,
                        name=self.series.name,
                    ),
               ]
    
        fig = go.Figure(data=data, layout=layout)
    
    
        # add mean, UCL and LCL line
        fig.add_shape(
                type="line",
                xref="paper",
                yref="y",
                x0=0,
                y0=self.mean,
                x1=1,
                y1=self.mean,
                line=dict(
                    color=f'rgba{defaults.black+(1,)}',
                    width=1,
                    dash='dash'
                ),
            )
        fig.add_annotation(
            dict(
                xref="paper",
                yref="y",
                x=1,
                y=self.mean,
                text=f'mean = {self.mean:1.2f}',
                showarrow=False,
                align='right',
                bgcolor=f'rgba{defaults.white+(0.5,)}',
            )
        )
        fig.add_shape(
                type="line",
                xref="paper",
                yref="y",
                x0=0,
                y0=self.ucl,
                x1=1,
                y1=self.ucl,
                line=dict(
                    color=f'rgba{defaults.black+(1,)}',
                    width=0.5,
                    dash='dash'
                ),
            )
        fig.add_annotation(
            dict(
                xref="paper",
                yref="y",
                x=1,
                y=self.ucl,
                text=f'UCL = {self.ucl:1.2f}',
                showarrow=False,
                align='right',
                bgcolor=f'rgba{defaults.white+(0.5,)}',
            )
        )
        fig.add_shape(
                type="line",
                xref="paper",
                yref="y",
                x0=0,
                y0=self.lcl,
                x1=1,
                y1=self.lcl,
                line=dict(
                    color=f'rgba{defaults.black+(1,)}',
                    width=0.5,
                    dash='dash'
                ),
            )
        fig.add_annotation(
            dict(
                xref="paper",
                yref="y",
                x=1,
                y=self.lcl,
                text=f'LCL = {self.lcl:1.2f}',
                showarrow=False,
                align='right',
                bgcolor=f'rgba{defaults.white+(0.5,)}',
            )
        )
        # x-ticks to nice datetime format
        if isinstance(self.series.index, pd.DatetimeIndex):
            fig.update_layout(xaxis_tickformat = '%d-%m-%Y')
        
        # set width en height if any

        if 'width' in kwargs and 'height' in kwargs:
            fig.update_layout(width=kwargs.get('width'), height=kwargs.get('height'))

        return fig

@bluebelt.core.decorators.class_methods
class RunChart():
    def __init__(self, series, alpha=0.05):
        
        # check arguments
        if not isinstance(series, pd.Series):
            raise ValueError('series is not a Pandas Series')
        
        self.series = series
        self.alpha = alpha
        
        self.calculate()
        
    def __str__(self):
        
        str_runs_about="runs about the median:"
        str_expected_runs_about="expected runs about the median:"
        str_longest_run_about="longest run about the median:"
        str_clustering=f"clustering (p ≈ {self.p_value_clustering:1.2f}):"
        str_mixtures=f"mixtures (p ≈ {self.p_value_mixtures:1.2f}):"
        
        str_runs_up_or_down="runs up or down:"
        str_expected_runs_up_or_down="expected runs up or down:"
        str_longest_run_up_or_down="longest run up or down:"
        str_trends=f"trends (p ≈ {self.p_value_trends:1.2f}):"
        str_oscillation=f"oscillation (p ≈ {self.p_value_oscillation:1.2f}):"

        fill = 32
        return (f'{str_runs_about:{fill}}{self.runs_about:1.0f}\n' +
                f'{str_expected_runs_about:{fill}}{self.expected_runs_about:1.0f}\n' +
                f'{str_longest_run_about:{fill}}{self.longest_run_about:1.0f}\n' +
                f'{str_clustering:{fill}}{self.clustering}\n' +
                f'{str_mixtures:{fill}}{self.mixtures}\n' +
                f'\n' +
                f'{str_runs_up_or_down:{fill}}{self.runs_up_or_down:1.0f}\n' +
                f'{str_expected_runs_up_or_down:{fill}}{self.expected_runs_up_or_down:1.0f}\n' +
                f'{str_longest_run_up_or_down:{fill}}{self.longest_run_up_or_down:1.0f}\n' +
                f'{str_trends:{fill}}{self.trends}\n' +
                f'{str_oscillation:{fill}}{self.oscillation}')


    def __repr__(self):
        return (f'{self.__class__.__name__}(runs_about={self.runs_about:1.0f}, expected_runs_about={self.expected_runs_about:1.0f}, longest_run_about={self.longest_run_about:1.0f}, runs_up_or_down={self.runs_up_or_down:1.0f}, expected_runs_up_or_down={self.expected_runs_up_or_down:1.0f}, longest_run_up_or_down={self.longest_run_up_or_down:1.0f}, p_value_clustering={self.p_value_clustering:1.2f}, p_value_mixtures={self.p_value_mixtures:1.2f}, p_value_trends={self.p_value_trends:1.2f}, p_value_oscillation={self.p_value_oscillation:1.2f}, clustering={self.clustering}, mixtures={self.mixtures}, trends={self.trends}, oscillation={self.oscillation})')

    def metrics(self):
        str_runs_about="runs about the median:"
        str_expected_runs_about="expected runs about the median:"
        str_longest_run_about="longest run about the median:"
        str_clustering=f"clustering (p ≈ {self.p_value_clustering:1.2f}):"
        str_mixtures=f"mixtures (p ≈ {self.p_value_mixtures:1.2f}):"
        
        str_runs_up_or_down="runs up or down:"
        str_expected_runs_up_or_down="expected runs up or down:"
        str_longest_run_up_or_down="longest run up or down:"
        str_trends=f"trends (p ≈ {self.p_value_trends:1.2f}):"
        str_oscillation=f"oscillation (p ≈ {self.p_value_oscillation:1.2f}):"

        fill = 32
        print( (f'{str_runs_about:{fill}}{self.runs_about:1.0f}\n' +
                f'{str_expected_runs_about:{fill}}{self.expected_runs_about:1.0f}\n' +
                f'{str_longest_run_about:{fill}}{self.longest_run_about:1.0f}\n' +
                f'{str_clustering:{fill}}{self.clustering}\n' +
                f'{str_mixtures:{fill}}{self.mixtures}\n' +
                f'\n' +
                f'{str_runs_up_or_down:{fill}}{self.runs_up_or_down:1.0f}\n' +
                f'{str_expected_runs_up_or_down:{fill}}{self.expected_runs_up_or_down:1.0f}\n' +
                f'{str_longest_run_up_or_down:{fill}}{self.longest_run_up_or_down:1.0f}\n' +
                f'{str_trends:{fill}}{self.trends}\n' +
                f'{str_oscillation:{fill}}{self.oscillation}'))
    
    def calculate(self):
        '''
        The number of runs about the median is the total number of runs above the median and 
        the total number of runs below the median.
        A run about the median is one or more consecutive points on the same side of the center line.
        A run ends when the line that connects the points crosses the center line.
        A new run begins with the next plotted point. 
        A data point equal to the median belongs to the run below the median.

        The number of runs up or down is the total count of upward and downward runs in the series.
        A run up or down ends when the direction changes.


        Clustering, mixtures, trends and oscillation
        A p-value that is less than the specified level of significance indicates clustering, mixtures, trends and/or oscillation
        '''

        median = self.series.median()

        longest_runs_about = [] #pd.Series(dtype=object)[
        longest_runs_up_or_down = [] #pd.Series(dtype=object)

        # runs

        for index, value in self.series.iteritems():

            # runs about the median
            if index == self.series.index[0]: # set above and start the first run
                above = True if value > median else False
                longest_run_about = 1
                run_about_length = 1
                runs_about = 0
            elif (value > median and not above) or (value <= median and above): # new run about
                runs_about += 1 # add an extra run
                above = not above # toggle the above value
                if run_about_length > longest_run_about:
                    longest_run_about = run_about_length
                    longest_runs_about = [self.series.loc[:index].iloc[-(longest_run_about+1):-1]]
                elif run_about_length == longest_run_about:
                    longest_runs_about += [self.series.loc[:index].iloc[-(longest_run_about+1):-1]]
                #longest_run_about = max(longest_run_about, run_about_length)
                run_about_length = 1
            elif index == self.series.index[-1]: # the last value might bring a longest run
                run_about_length += 1
                if run_about_length > longest_run_about:
                    longest_run_about = run_about_length
                    longest_runs_about = [self.series.loc[:index].iloc[-(longest_run_about):]]
                elif run_about_length == longest_run_about:
                    longest_runs_about += [self.series.loc[:index].iloc[-(longest_run_about):]]
            else:
                run_about_length += 1

            # runs up or down
            if index == self.series.index[0]: # set the first value
                previous_value = value
            elif index == self.series.index[1]: # set up and start first run
                up = True if value > previous_value else False
                longest_run_up_or_down = 1
                run_up_or_down_length = 1
                runs_up_or_down = 1
                previous_value = value

            elif (value > previous_value and not up) or (value <= previous_value and up): # new run up or down
                runs_up_or_down += 1 # add an extra run
                up = not up # toggle up
                if run_up_or_down_length > longest_run_up_or_down:
                    longest_run_up_or_down = run_up_or_down_length
                    longest_runs_up_or_down = [self.series.loc[:index].iloc[-(longest_run_up_or_down+1):-1]]
                elif run_up_or_down_length == longest_run_up_or_down:
                    longest_runs_up_or_down += [self.series.loc[:index].iloc[-(longest_run_up_or_down+1):-1]]   
                run_up_or_down_length = 1
                previous_value = value

            elif index == self.series.index[-1]: # the last value might bring a longest run
                run_up_or_down_length += 1
                if run_up_or_down_length > longest_run_up_or_down:
                    longest_run_up_or_down = run_up_or_down_length
                    longest_runs_up_or_down = [self.series.loc[:index].iloc[-(longest_run_up_or_down):]]
                    
                elif run_up_or_down_length == longest_run_up_or_down:
                    longest_runs_up_or_down += [self.series.loc[:index].iloc[-(longest_run_up_or_down):]]

            else:
                run_up_or_down_length += 1
                previous_value = value



        # expected runs
        m = self.series[self.series > self.series.median()].count()
        n = self.series[self.series <= self.series.median()].count()
        N = self.series.count()

        expected_runs_about = 1 + (2 * m * n) / N

        expected_runs_up_or_down = (2 * (m + n) - 1) / 3

        # clustering and mixtures
        p_value_clustering = stats.norm.cdf((runs_about - 1 - ((2 * m * n) / N)) / (((2 * m * n * (2 * m * n - N)) / (N**2 * (N - 1)))**0.5))
        p_value_mixtures = 1 - p_value_clustering

        clustering = True if p_value_clustering < self.alpha else False
        mixtures = True if p_value_mixtures < self.alpha else False

        # trends and oscillation
        p_value_trends = stats.norm.cdf((runs_up_or_down - (2 * N - 1) / 3) / ((16 * N - 29) / 90)**0.5)
        p_value_oscillation = 1 - p_value_trends

        trends = True if p_value_trends < self.alpha else False
        oscillation = True if p_value_oscillation < self.alpha else False
        
        self.runs_about = runs_about
        self.expected_runs_about = expected_runs_about
        self.longest_run_about = longest_run_about
        self.runs_up_or_down = runs_up_or_down
        self.expected_runs_up_or_down = expected_runs_up_or_down
        self.longest_run_up_or_down = longest_run_up_or_down
        self.p_value_clustering = p_value_clustering
        self.p_value_mixtures = p_value_mixtures
        self.p_value_trends = p_value_trends
        self.p_value_oscillation = p_value_oscillation
        self.clustering = clustering
        self.mixtures = mixtures
        self.trends = trends
        self.oscillation = oscillation
        self.longest_runs_about = longest_runs_about
        self.longest_runs_up_or_down = longest_runs_up_or_down
        
    def plot(self, **kwargs):
        
        style = kwargs.get('style', bluebelt.styles.paper)
        kwargs.pop('style', None)

        path = kwargs.pop('path', None)        
        
        fig, ax = plt.subplots(nrows=1, ncols=1, **kwargs)
        
        # observations
        ax.plot(self.series, **style.run_chart.plot)
        

        # longest run(s) about the median and longest run(s) up or down
        ylim = ax.get_ylim() # get ylim to set it back later

        for run in self.longest_runs_about:
            ax.fill_between(run.index, run.values, ylim[0], **style.run_chart.longest_runs_about_fill_between)

        for run in self.longest_runs_up_or_down:
            ax.fill_between(run.index, run.values, ylim[1], **style.run_chart.longest_runs_up_or_down_fill_between)
        
        ax.set_ylim(ylim[0], ylim[1]) # reset ylim

        # mean
        ax.axhline(self.series.median(), zorder=1, **style.run_chart.median_axhline)
        ax.text(ax.get_xlim()[1], self.series.median(), f' median = {self.series.median():1.2f}', **style.run_chart.median_text)

        ax.text(ax.get_xlim()[1], ylim[0], f' longest {"run" if len(self.longest_runs_about)==1 else "runs"}\n about the\n median = {self.longest_run_about}', **style.run_chart.longest_runs_about_text)
        ax.text(ax.get_xlim()[1], ylim[1], f' longest {"run" if len(self.longest_runs_up_or_down)==1 else "runs"}\n up or down = {self.longest_run_up_or_down}', **style.run_chart.longest_runs_up_or_down_text)

        # labels
        ax.set_title(f'run chart of {self.series.name}', **style.run_chart.title)
        #ax.set_xlabel(self.series.index.name)
        ax.set_ylabel('value')

        # set x axis locator
        mpl_format.axisformat(ax, self.series)

        plt.gcf().subplots_adjust(right=0.8)

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

        
    def plotly(self, **kwargs):
        
        layout = go.Layout(
            title=f'run chart of {self.series.name}',    
            plot_bgcolor=f'rgba{defaults.white+(1,)}',
            hovermode="closest",
            xaxis=dict(
                title=self.series.index.name,
                linecolor=f'rgba{defaults.black+(1,)}',
                zeroline=False,
                ticks="outside",
                tickwidth=1,
                tickcolor=f'rgba{defaults.black+(1,)}',
                ticklen=5,
                showgrid=True,
                gridwidth=0.5,
                gridcolor=f'rgba{defaults.light_grey+(0.2,)}',
                mirror=True,
            ),
            yaxis=dict(
                title="value",
                linecolor=f'rgba{defaults.black+(1,)}',
                zeroline=False,
                ticks="outside",
                tickwidth=1,
                tickcolor=f'rgba{defaults.black+(1,)}',
                ticklen=5,
                showgrid=True,
                gridwidth=0.5,
                gridcolor=f'rgba{defaults.light_grey+(1,)}',
                mirror=True,
            ),
        )
        
        data = []
        
        for idx, trace in enumerate(self.longest_runs_about):
            data.append(go.Scatter(
                x=trace.index,
                y=trace.values,
                line=dict(
                    width=9,
                    color=f'rgba{defaults.red+(0.4,)}',
                ),
                mode='lines',
                name=f'longest {"run" if len(self.longest_runs_about)==1 else "runs"} about the median ({self.longest_run_about})',
                legendgroup="runs_about",
                showlegend=True if idx==0 else False,
            ))

        for idx, trace in enumerate(self.longest_runs_up_or_down):
            data.append(go.Scatter(
                x=trace.index,
                y=trace.values,
                line=dict(
                    width=9,
                    color=f'rgba{defaults.blue+(0.4,)}',
                ),
                mode='lines',
                name=f'longest {"run" if len(self.longest_runs_up_or_down)==1 else "runs"} up or down ({self.longest_run_up_or_down})',
                legendgroup="runs_up_down",
                showlegend=True if idx==0 else False,
            ))


        data.append(go.Scatter(
            x=self.series.index,
            y=self.series.values,
            line=dict(
                    width=1,
                    color=f'rgba{defaults.blue+(1,)}',
                ),
            marker=dict(
                color=f'rgba{defaults.black+(1,)}',
                size=5,
            ),
            mode='lines+markers',
            showlegend=False,
        ))
        
        fig = go.Figure(data=data, layout=layout)

        # legend position
        fig.update_layout(legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.02,
            xanchor="right",
            x=1
        ))

    
        # add median line
        fig.add_shape(
                type="line",
                xref="paper",
                yref="y",
                x0=0,
                y0=self.series.median(),
                x1=1,
                y1=self.series.median(),
                line=dict(
                    color=f'rgba{defaults.grey+(1,)}',
                    width=1,
                    dash='dash'
                ),
            )
        fig.add_annotation(
            dict(
                xref="paper",
                yref="y",
                x=1,
                y=self.series.median(),
                text=f'median = {self.series.median():1.2f}',
                showarrow=False,
                align='right',
                bgcolor=f'rgba{defaults.white+(0.5,)}',
            )
        )
        
        # x-ticks to nice datetime format
        if isinstance(self.series.index, pd.DatetimeIndex):
            fig.update_layout(xaxis_tickformat = '%d-%m-%Y')
        
        # set width en height if any
        if 'width' in kwargs and 'height' in kwargs:
            fig.update_layout(width=kwargs.get('width'), height=kwargs.get('height'))

        return fig

@bluebelt.core.decorators.class_methods
class ProcessCapability():
    def __init__(self,
                 series,
                 target=None,
                 usl=None,
                 ub=None,
                 lsl=None,
                 lb=None,
                 subgroups=None,
                 subgroup_size=1,
                 tolerance=6,
                ):
        
        # check arguments
        if not isinstance(series, pd.Series):
            raise ValueError('series is not a Pandas Series')
        
        self.series = series
        self.target = target
        self.usl = usl
        self.lsl = lsl
        self.ub = ub
        self.lb = lb
        self.subgroups = subgroups
        self.subgroup_size = subgroup_size
        self.tolerance = tolerance
        
        # check parameters
        self.check()
        
        self.calculate()
    def test(self):
        return np.power(self.series - self.target, 2)

    def check(self):
        # check if all parameters are ok

        # limits and bounds
        if (self.lb and self.lsl) and self.lb != self.lsl:
            raise ValueError("You can specify a lower bound (lb) or a lower specification limit (lsl) but not both.")
        if (self.ub and self.usl) and self.ub != self.usl:
            raise ValueError("You can specify a upper bound (ub) or a upper specification limit (usl) but not both.")
        
    def calculate(self):
        
        # basic statistics
        self.min = self.series.min()
        self.max = self.series.max()
        self.mean = self.series.mean()
        self.std = self.series.std()
        self.subgroups = get_subgroups(self.series, subgroups=self.subgroups, subgroup_size=self.subgroup_size)
        self.std_within = bluebelt.statistics.std.StdWithin(self.subgroups).std
        self.size = self.series.size

        self._lsl = self.lb if self.lb is not None else self.lsl if self.lsl is not None else None
        self._usl = self.ub if self.ub is not None else self.usl if self.usl is not None else None
        
        self._lsl_or_min = self.lb if self.lb is not None else self.lsl if self.lsl is not None else self.min
        self._usl_or_max = self.ub if self.ub is not None else self.usl if self.usl is not None else self.max
        
        if (self._usl is not None) and (self._lsl is not None):
            self._midpoint = (self._usl + self._lsl) / 2
        else:
            self._midpoint = self.mean

        # performance
        self.observed_lt_lsl = self.get_observed_lt_lsl()
        self.observed_gt_usl = self.get_observed_gt_usl()
        self.observed_performance = self.get_observed_performance()

        self.expected_lt_lsl_within = self.get_expected_lt_lsl_within()
        self.expected_gt_usl_within = self.get_expected_gt_usl_within()
        self.expected_performance_within = self.get_expected_performance_within()

        self.expected_lt_lsl_overall = self.get_expected_lt_lsl_overall()
        self.expected_gt_usl_overall = self.get_expected_gt_usl_overall()
        self.expected_performance_overall = self.get_expected_performance_overall()
        
        # within capability
        self.cp = self.get_cp()
        self.cpl = self.get_cpl()
        self.cpu = self.get_cpu()
        self.cpk = self.get_cpk()
        self.ccpk = self.get_ccpk()
        
        # overall capability
        self.pp = self.get_pp()
        self.ppl = self.get_ppl()
        self.ppu = self.get_ppu()
        self.ppk = self.get_ppk()
        self.cpm = self.get_cpm()
    
    # performance
    def get_observed_lt_lsl(self):
        return (self.series[self.series < self._lsl].count() / self.size) * 1000000 if self._lsl is not None else 0

    def get_observed_gt_usl(self):
        return (self.series[self.series > self._usl].count() / self.size) * 1000000 if self._usl is not None else 0

    def get_observed_performance(self):
        return self.get_observed_lt_lsl() + self.get_observed_gt_usl()

    def get_expected_lt_lsl_within(self):
        return (1 - stats.norm.cdf((self.mean - self._lsl) / self.std_within, loc=0, scale=1)) * 1000000 if self._lsl is not None else 0

    def get_expected_gt_usl_within(self):
        return (1 - stats.norm.cdf((self._usl - self.mean) / self.std_within, loc=0, scale=1)) * 1000000 if self._usl is not None else 0

    def get_expected_performance_within(self):
        return self.get_expected_lt_lsl_within() + self.get_expected_gt_usl_within()

    def get_expected_lt_lsl_overall(self):
        return (1 - stats.norm.cdf((self.mean - self._lsl) / self.std, loc=0, scale=1)) * 1000000 if self._lsl is not None else 0

    def get_expected_gt_usl_overall(self):
        return (1 - stats.norm.cdf((self._usl - self.mean) / self.std, loc=0, scale=1)) * 1000000 if self._usl is not None else 0

    def get_expected_performance_overall(self):
        return self.get_expected_lt_lsl_overall() + self.get_expected_gt_usl_overall()

    # within capability
    def get_cp(self):
        """
        Cp is a measure of the potential capability of the process. It is calculated by taking the ratio of the specification spread (USL – LSL)
        and the process spread (the tolerance * sigma variation) based on the standard deviation within subgroups.

        Cp = (usl - lsl) / (tolerance * std_within)
        """
        if (self._usl is not None) and (self._lsl is not None):
            return (self._usl - self._lsl) / (self.tolerance * self.std_within)
        else:
            return None

    def get_cpl(self):
        """
        Cpl = (mean - lsl) / ((tolerance / 2) * std_within)
        """
        if self._lsl is not None:
            return (self.mean - self._lsl) / (self.tolerance * 0.5 * self.std_within)
        else:
            return None

    def get_cpu(self):
        """
        Cpu = (usl - mean) / ((tolerance / 2) * std_within)
        """
        if self._usl is not None:
            return (self._usl - self.mean) / (self.tolerance * 0.5 * self.std_within)
        else:
            return None

    def get_cpk(self):
        if (self._usl is not None) and (self._lsl is not None):
            return min(self.get_cpl(), self.get_cpu())
        else:
            return None
    
    def get_ccpk(self):
        # calculate mu
        if self.target is not None:
            mu = self.target
        elif (self._usl is not None) and (self._lsl is not None):
            mu = (self._lsl + self._usl) / 2
        else:
            mu = self.mean

        # calculate ccpk
        if (self._usl is not None) and (self._lsl is not None):
            ccpk = min((self._usl - mu), (mu - self._lsl)) / ((self.tolerance / 2) * self.std_within)
        elif self._usl is not None:
            ccpk = (self._usl - mu) / ((self.tolerance / 2) * self.std_within) 
        elif self._lsl is not None:
            ccpk = (mu - self._lsl) / ((self.tolerance / 2) * self.std_within) 
        else:
            ccpk = None

        return ccpk
    
    # overall capability
    def get_pp(self):
        # Pp = (USL – LSL) / tolerance * sigma
        if (self._usl is not None) and (self._lsl is not None):
            return (self._usl - self._lsl) / (self.tolerance * self.std)
        else:
            return None

    def get_ppl(self):
        if self._lsl is not None:
            return (self.mean - self._lsl) / ((self.tolerance / 2) * self.std)
        else:
            return None

    def get_ppu(self):
        if self._usl is not None:
            return (self._usl - self.mean) / ((self.tolerance / 2) * self.std)
        else:
            return None

    def get_ppk(self):
        if (self._usl is not None) and (self._lsl is not None):
            return min(self.get_ppl(), self.get_ppu())
        else:
            return None

    def get_cpm(self):
        if not self.target:
            return None
        
        elif self._lsl or self._usl:
            if not self._usl:
                # LSL and target only
                numerator = self.target - (self.lb or self.lsl)
                denominator_factor = 0.5
            elif not self._lsl:
                # USL and target only
                numerator = self._usl - self.target
                denominator_factor = 0.5
            elif self.target == (self._lsl + self._usl) / 2:
                numerator = self._usl - self._lsl
                denominator_factor = 1
            else:
                numerator = min(self.target - self._lsl, self._usl - self.target)
                denominator_factor = 0.5
        else:
            return None

        # get subgroups
        subgroups = get_subgroups(self.series, subgroup_size=5)

        denominator = self.tolerance * ((sum([sum((subgroups[col].dropna()-self.target)**2) for col in subgroups.columns]) / self.series.size)**0.5)

        return numerator / (denominator_factor * denominator)
    
    def __str__(self):
        
        fill = 15
        
        # process data
        digits = f'{bluebelt.helpers.text.format_digits([self.lsl, self.usl, self.target, self.mean, self.std_within, self.std], 6)}'

        str_target = f'{"target":{fill}}{self.target}'
        str_lsl = f'{"LB" if self.lb else "LSL":{fill}}{self._lsl:{digits}}' if self._lsl is not None else ''
        str_usl = f'{"UB" if self.ub else "USL":{fill}}{self._usl:{digits}}' if self._usl is not None else ''
        str_mean = f'{"mean":{fill}}{self.mean:{digits}}'
        str_n = f'{"n":{fill}}{self.size}'
        str_std_within = f'{"std within":{fill}}{self.std_within:{digits}}'
        str_std_overall = f'{"std overall":{fill}}{self.std:{digits}}'
        
        # within capability
        str_cp = f'{"Cp":{fill}}{self.cp:1.2f}' if self.cp is not None else f'{"Cp":{fill}}*'
        str_cpl = f'{"Cpl":{fill}}{self.cpl:1.2f}' if self.cpl is not None else f'{"Cpl":{fill}}*'
        str_cpu = f'{"Cpu":{fill}}{self.cpu:1.2f}' if self.cpu is not None else f'{"Cpu":{fill}}*'
        str_cpk = f'{"Cpk":{fill}}{self.cpk:1.2f}' if self.cpk is not None else f'{"Cpk":{fill}}*'
        str_ccpk = f'{"CCpk":{fill}}{self.ccpk:1.2f}' if self.ccpk is not None else f'{"CCpk":{fill}}*'

        # overall capability
        str_pp = f'{"Pp":{fill}}{self.pp:1.2f}' if self.pp is not None else f'{"Pp":{fill}}*'
        str_ppl = f'{"Ppl":{fill}}{self.ppl:1.2f}' if self.ppl is not None else f'{"Ppl":{fill}}*'
        str_ppu = f'{"Ppu":{fill}}{self.ppu:1.2f}' if self.ppu is not None else f'{"Ppu":{fill}}*'
        str_ppk = f'{"Ppk":{fill}}{self.ppk:1.2f}' if self.ppk is not None else f'{"Ppk":{fill}}*'
        str_cpm = f'{"Cpm":{fill}}{self.cpm:1.2f}' if self.cpm is not None else f'{"Cpm":{fill}}*'

        # performance
        str_observed_lt_lsl = f'{"PPM < LSL":{fill}}{self.observed_lt_lsl:1.0f}'
        str_observed_gt_usl = f'{"PPM > USL":{fill}}{self.observed_gt_usl:1.0f}'
        str_observed_performance = f'{"PPM":{fill}}{self.observed_performance:1.0f} ({self.observed_performance / 10000:1.2f}%)'
        
        str_expected_lt_lsl_within = f'{"PPM < LSL":{fill}}{self.expected_lt_lsl_within:1.0f}'
        str_expected_gt_usl_within = f'{"PPM > USL":{fill}}{self.expected_lt_lsl_within:1.0f}'
        str_expected_performance_within = f'{"PPM":{fill}}{self.expected_performance_within:1.0f} ({self.expected_performance_within / 10000:1.2f}%)'
        
        str_expected_lt_lsl_overall = f'{"PPM < LSL":{fill}}{self.expected_lt_lsl_overall:1.0f}'
        str_expected_gt_usl_overall = f'{"PPM > USL":{fill}}{self.expected_gt_usl_overall:1.0f}'
        str_expected_performance_overall = f'{"PPM":{fill}}{self.expected_performance_overall:1.0f} ({self.expected_performance_overall / 10000:1.2f}%)'
        
        width = 35

        result = (
            f'{"Process Data":{width}}{"Potential Capability":{width}}{"Overall Capability":{width}}\n' +
            f'{str_target:{width}}' + f'{str_cp:{width}}' + f'{str_pp:{width}}' + '\n' +
            f'{str_lsl:{width}}' + f'{str_cpl:{width}}' + f'{str_ppl:{width}}' + '\n' +
            f'{str_usl:{width}}' + f'{str_cpu:{width}}' + f'{str_ppu:{width}}' + '\n' +
            f'{str_mean:{width}}' + f'{str_cpk:{width}}' + f'{str_ppk:{width}}' + '\n' +
            f'{str_n:{width}}' + f'{str_ccpk:{width}}' + f'{str_cpm:{width}}' + '\n' +
            f'{str_std_within:{width}}' + '\n' +
            f'{str_std_overall:{width}}' + '\n' + '\n' +
            f'{"Observed Performance":{width}}{"Expected Performance (Within)":{width}}{"Expected Performance (Overall)":{width}}\n' +
            f'{str_observed_lt_lsl:{width}}' + f'{str_expected_lt_lsl_within:{width}}' + f'{str_expected_lt_lsl_overall:{width}}' + '\n' +
            f'{str_observed_gt_usl:{width}}' + f'{str_expected_gt_usl_within:{width}}' + f'{str_expected_gt_usl_overall:{width}}' + '\n' +
            f'{str_observed_performance:{width}}' + f'{str_expected_performance_within:{width}}' + f'{str_expected_performance_overall:{width}}'
            )
                
        return result

    def __repr__(self):
        target_value = self.target or 'None'

        lsl_text = 'lb' if self.lb is not None else 'lsl'
        lsl_value = self._lsl or 'None'

        usl_text = 'ub' if self.ub is not None else 'usl'
        usl_value = self._usl or 'None'

        
        return (f'{self.__class__.__name__}(n={self.size}, target={target_value}, {lsl_text}={lsl_value}, {usl_text}={usl_value})')
    
    @property
    def result(self):
        print(self)

    def df(self):
        df_md = pd.DataFrame({
            'metric': ["target", "LB" if self.lb else "LSL", "UB" if self.ub else "USL", "% < LSL", "% > USL", "Observed Performance", "Pp", "Ppk"],
            'value': [self.target, self._lsl, self._usl, self.observed_lt_lsl, self.observed_gt_usl, self.observed_performance, self.pp, self.ppk],
            }
        )
        
        return df_md

    def md(self):
        print(self.df().to_markdown(index=False))

        
    def plot(self, **kwargs):
        
        style = kwargs.get('style', bluebelt.styles.paper)
        kwargs.pop('style', None)

        path = kwargs.pop('path', None)

        # get bins
        min_bins = kwargs.pop('min_bins', 10)
        max_bins = kwargs.pop('max_bins', 20)

        histogram_points = [x for x in [self._lsl_or_min, self._usl_or_max, self.target] if x is not None]
        bins = bluebelt.helpers.histogram.bins(series=self.series, points=histogram_points, min_bins=min_bins, max_bins=max_bins)

        def _set_patch_style(patch, style):
            for key in ['facecolor', 'edgecolor', 'linewidth', 'hatch', 'fill']:
                if key in style:
                    eval(f'patch.set_{key}(style.get(key))')
        
        fig, ax = plt.subplots(nrows=1, ncols=1, **kwargs)
        
        # 1. histogram ############################################
        n, bins, patches = ax.hist(self.series, bins=bins, **style.process_capability.histogram)

        for patch in patches:
            # catch patches 

            # LSL
            if self._lsl is not None:

                # < LSL
                if patch.get_x() + patch.get_width() <= self._lsl:
                    patch.set_fill(False)
                    patch.set_hatch('')
                    patch_copy = copy.copy(patch)
                    _set_patch_style(patch_copy, style.process_capability.histogram_fill_out_of_range)
                    ax.add_patch(patch_copy)

                # on LSL
                elif patch.get_x() < self._lsl and patch.get_x() + patch.get_width() > self._lsl:
                    # split patch
                    patch.set_fill(False)
                    patch.set_hatch('')
                    # first half
                    patch_width_1 = self._lsl - patch.get_x()
                    patch_copy = copy.copy(patch)
                    _set_patch_style(patch_copy, style.process_capability.histogram_fill_out_of_range)
                    patch_copy.set_width(patch_width_1)
                    ax.add_patch(patch_copy)
                
                    # second half
                    patch_width_2 = (patch.get_x()+patch.get_width()) - self._lsl
                    patch_copy = copy.copy(patch)
                    _set_patch_style(patch_copy, style.process_capability.histogram_fill_in_range)
                    patch_copy.set_width(patch_width_2)
                    patch_copy.set_x(patch.get_x()+patch_width_1)
                    ax.add_patch(patch_copy)
            
            if self._usl is not None:
                # > USL
                if patch.get_x() >= self._usl:
                    patch.set_fill(False)
                    patch.set_hatch('')
                    patch_copy = copy.copy(patch)
                    _set_patch_style(patch_copy, style.process_capability.histogram_fill_out_of_range)
                    ax.add_patch(patch_copy)

                # on USL
                elif patch.get_x() <= self._usl and patch.get_x() + patch.get_width() > self._usl:
                    # split patch
                    patch.set_fill(False)
                    patch.set_hatch('')
                    # first half
                    patch_width_1 = self._usl - patch.get_x()
                    patch_copy = copy.copy(patch)
                    _set_patch_style(patch_copy, style.process_capability.histogram_fill_in_range)
                    patch_copy.set_width(patch_width_1)
                    ax.add_patch(patch_copy)
                    
                    # second half
                    patch_width_2 = (patch.get_x()+patch.get_width()) - self._usl
                    patch_copy = copy.copy(patch)
                    _set_patch_style(patch_copy, style.process_capability.histogram_fill_out_of_range)
                    patch_copy.set_width(patch_width_2)
                    patch_copy.set_x(patch.get_x()+patch_width_1)
                    ax.add_patch(patch_copy)

        # get current limits
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()

        # fit a normal distribution to the data
        pdf_x = np.linspace(xlims[0], xlims[1], 100)

        # normal plot overall
        ax.plot(pdf_x, stats.norm.pdf(pdf_x, self.mean, self.std), label='overall', **style.process_capability.normal_plot_overall)
        
        # normal plot within
        ax.plot(pdf_x, stats.norm.pdf(pdf_x, self.mean, self.std_within), label='within', **style.process_capability.normal_plot_within)
        
        # target
        if self.target is not None:
            ax.axvline(x=self.target, ymin=0, ymax=1, **style.process_capability.target_axvline)
            ax.text(self.target, ylims[1]*0.9, f'target', **style.process_capability.sl_text)

        # LSL, USL
        if self._lsl is not None:
            ax.axvline(x=self._lsl, ymin=0, ymax=1, **style.process_capability.sl_axvline)
            lsl_text = 'LB' if self.lb is not None else 'LSL'
            ax.text(self._lsl, ylims[1]*0.9, lsl_text, **style.process_capability.sl_text)
        
        if self._usl is not None:
            ax.axvline(x=self._usl, ymin=0, ymax=1, **style.process_capability.sl_axvline)
            usl_text = 'UB' if self.ub is not None else 'USL'
            ax.text(self._usl, ylims[1]*0.9, usl_text, **style.process_capability.sl_text)

        # change xlim if needed
        xlims_min = min(self.min, self._lsl_or_min, self._usl_or_max, (self.target or self.mean))
        xlims_max = max(self.max, self._lsl_or_min, self._usl_or_max, (self.target or self.mean))
        xlims_margin = (xlims_max - xlims_min) * plt.rcParams['axes.xmargin']
        xlims = (xlims_min - xlims_margin, xlims_max + xlims_margin)

        # reset limits
        ax.set_xlim(xlims)
        ax.set_ylim(ylims)

        # set ticks
        ax.set_yticks([])
        
        # labels
        ax.set_title(f'process capability analysis of {self.series.name}', **style.process_capability.title)
        #ax.set_xlabel(self.series.index.name)
        ax.legend(loc='upper right')

        plt.gcf().subplots_adjust(right=0.8)

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig



import bluebelt.helpers.defaults as defaults

# main plot | pyplot.plot
plot = {
    'marker': 'o',
    'zorder': 50, 
}

# main plot background (used to create space in UCL and LCL area) | pyplot.plot
plot_background = {
    'marker': None,
    'color': defaults.white,
    'linewidth': 11,
    'zorder': 30, 
}

# ucl axhline | pyplot.axhline
ucl_axhline = {
    'color': defaults.grey,
    'linestyle': 'dashed',
    'linewidth': 1,
    'zorder': 1, 
}

# lcl axhline | pyplot.axhline
lcl_axhline = {
    'color': defaults.grey,
    'linestyle': 'dashed',
    'linewidth': 1,
    'zorder': 1, 
}

# mean axhline | pyplot.axhline
mean_axhline = {
    'color': defaults.grey,
    'linestyle': 'dashed',
    'linewidth': 1,
    'zorder': 40, 
    
}

# ucl fill_between | pyplot.fill_between
ucl_fill_between = {
    'color': None,
    'hatch': '//',
    'linestyle': 'solid',
    'linewidth': 0,
    'zorder': 10, 
}

# lcl fill_between | pyplot.fill_between
lcl_fill_between = {
    'color': None,
    'hatch': '//',
    'linestyle': 'solid',
    'linewidth': 0,
    'zorder': 10, 
}

# outlier background | pyplot.plot
outlier_background = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 17,
    'markerfacecolor': defaults.white,
    'zorder': 40, 
}

# outlier marker | pyplot.plot
outlier = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 9,
    'markerfacecolor': defaults.white,
    'markeredgecolor': defaults.black+(1,),
    'markeredgewidth': 1,
    'zorder': 45, 
}

# mean text | pyplot.text
mean_text = {
    'ha': 'left',
    'va': 'center',
    'zorder': 50,
}

# ucl text | pyplot.text
ucl_text = {
    'ha': 'left',
    'va': 'center',
    'zorder': 50,
}

# lcl text | pyplot.text
lcl_text = {
    'ha': 'left',
    'va': 'center',
    'zorder': 50,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}



def subset(frame, inverse=False, **kwargs):
        
    #check if filter values are lists and repair

    # filters = {key: ([value] if not isinstance(value, list) else value) for key, value in kwargs.items()}
    filters={}
    for col in kwargs:
        if col in frame.columns:
            values = kwargs.get(col) if isinstance(kwargs.get(col), list) else [kwargs.get(col)]
            for value in values:
                if value not in frame[col].values:
                    raise ValueError(f'{value} is not in {col}')
            filters[col]=values
        else:
            raise ValueError(f'{col} is not in the dataframe')

    #filter  
    if inverse:
        frame=frame[frame.isin(filters).sum(axis=1) != len(filters.keys())]
    else:
        frame=frame[frame.isin(filters).sum(axis=1) == len(filters.keys())]
        
    return frame


import numpy as np
import pandas as pd
from decimal import Decimal

def bins(series, points, min_bins=10, max_bins=20):

    if not isinstance(series, pd.Series):
        raise ValueError('series is not a Pandas Series')

    if not isinstance(points, (tuple, list)):
        raise ValueError('points is not a tuple or a list')

    if not isinstance(min_bins, int):
        raise ValueError('min_bins is not an integer')
    else:
        if min_bins < 1:
            min_bins = 1
    
    if not isinstance(max_bins, int):
        raise ValueError('max_bins is not an integer')
    else:
        if max_bins < 5:
            max_bins = 5
    
    
    # decimal places for points
    decimal_places = -min([Decimal(str(x)).as_tuple().exponent for x in points])
    
    gcd_factor = 10**decimal_places
    bin_size = np.gcd.reduce([int(round(x*gcd_factor)) for x in points]) / gcd_factor
    
    # adjust bin_size
    series_range = series.max() - series.min()

    ## min_bins
    range_factor = np.ceil(min_bins/(series_range/bin_size))
    bin_size = bin_size / range_factor
    
    ## max_bins
    range_factor = np.ceil((series_range/bin_size)/max_bins)
    bin_size = bin_size * range_factor

    # new decimal places
    decimal_places = -Decimal(str(bin_size)).as_tuple().exponent
    
    # get start and end
    start = round(series.min() - (series.min() % bin_size), decimal_places)
    end = round(series.max() - (series.max() % bin_size) + bin_size, decimal_places)

    return np.arange(start, end, bin_size)